#
#   This file is part of m.css.
#
#   Copyright © 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024
#             Vladimír Vondruš <mosra@centrum.cz>
#
#   Permission is hereby granted, free of charge, to any person obtaining a
#   copy of this software and associated documentation files (the "Software"),
#   to deal in the Software without restriction, including without limitation
#   the rights to use, copy, modify, merge, publish, distribute, sublicense,
#   and/or sell copies of the Software, and to permit persons to whom the
#   Software is furnished to do so, subject to the following conditions:
#
#   The above copyright notice and this permission notice shall be included
#   in all copies or substantial portions of the Software.
#
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#   DEALINGS IN THE SOFTWARE.
#

from docutils.parsers import rst
from docutils.parsers.rst import directives
from docutils.parsers.rst.roles import set_classes
from docutils import nodes

class Transition(rst.Directive):
    final_argument_whitespace = True
    has_content = False
    required_arguments = 1

    def run(self):
        text = ' '.join(self.arguments)
        title_nodes, _ = self.state.inline_text(text, self.lineno)
        transition_node = nodes.transition('', *title_nodes)
        transition_node['classes'] += ['m-transition']
        return [transition_node]

class Note(rst.Directive):
    final_argument_whitespace = True
    has_content = True
    optional_arguments = 1
    option_spec = {'class': directives.class_option}

    style_class = ''

    def run(self):
        set_classes(self.options)

        if len(self.arguments) == 1:
            title_text = self.arguments[0]
            title_nodes, _ = self.state.inline_text(title_text, self.lineno)
            title_node = nodes.title('', '', *title_nodes)

        text = '\n'.join(self.content)
        topic_node = nodes.topic(text, **self.options)
        topic_node['classes'] += ['m-note', self.style_class]

        if len(self.arguments) == 1:
            topic_node.append(title_node)

        self.state.nested_parse(self.content, self.content_offset,
                                topic_node)
        return [topic_node]

class DefaultNote(Note):
    style_class = 'm-default'

class PrimaryNote(Note):
    style_class = 'm-primary'

class SuccessNote(Note):
    style_class = 'm-success'

class WarningNote(Note):
    style_class = 'm-warning'

class DangerNote(Note):
    style_class = 'm-danger'

class InfoNote(Note):
    style_class = 'm-info'

class DimNote(Note):
    style_class = 'm-dim'

class Block(rst.Directive):
    final_argument_whitespace = True
    has_content = True
    required_arguments = 1
    option_spec = {'class': directives.class_option}

    style_class = ''

    def run(self):
        set_classes(self.options)

        title_text = self.arguments[0]
        title_elements, _ = self.state.inline_text(title_text, self.lineno)
        title_node = nodes.title('', '', *title_elements)

        text = '\n'.join(self.content)
        topic_node = nodes.topic(text, **self.options)
        topic_node['classes'] += ['m-block', self.style_class]
        topic_node.append(title_node)

        self.state.nested_parse(self.content, self.content_offset,
                                topic_node)

        return [topic_node]

class DefaultBlock(Block):
    style_class = 'm-default'

class PrimaryBlock(Block):
    style_class = 'm-primary'

class SuccessBlock(Block):
    style_class = 'm-success'

class WarningBlock(Block):
    style_class = 'm-warning'

class DangerBlock(Block):
    style_class = 'm-danger'

class InfoBlock(Block):
    style_class = 'm-info'

class DimBlock(Block):
    style_class = 'm-dim'

class FlatBlock(Block):
    style_class = 'm-flat'

class Frame(rst.Directive):
    final_argument_whitespace = True
    has_content = True
    optional_arguments = 1
    option_spec = {'class': directives.class_option}

    style_class = ''

    def run(self):
        set_classes(self.options)

        if len(self.arguments) == 1:
            title_text = self.arguments[0]
            title_nodes, _ = self.state.inline_text(title_text, self.lineno)
            title_node = nodes.title('', '', *title_nodes)

        text = '\n'.join(self.content)
        topic_node = nodes.topic(text, **self.options)
        topic_node['classes'] += ['m-frame', self.style_class]

        if len(self.arguments) == 1:
            topic_node.append(title_node)

        self.state.nested_parse(self.content, self.content_offset,
                                topic_node)
        return [topic_node]

class _Figure(rst.Directive):
    final_argument_whitespace = True
    has_content = True
    optional_arguments = 1
    option_spec = {'class': directives.class_option}

    style_classes = []

    def run(self):
        set_classes(self.options)

        text = '\n'.join(self.content)
        figure_node = nodes.figure(text, **self.options)
        figure_node['classes'] += self.style_classes

        self.state.nested_parse(self.content, self.content_offset,
                                figure_node)

        # Insert the title node, if any, right after the code / math / graph.
        # There could be things like the class directive before, so be sure to
        # find the right node.
        if len(self.arguments) == 1:
            title_text = self.arguments[0]
            title_nodes, _ = self.state.inline_text(title_text, self.lineno)
            title_node = nodes.caption('', '', *title_nodes)

            for i, child in enumerate(figure_node):
                if isinstance(child, nodes.raw) or isinstance(child, nodes.literal_block):
                    figure_node.insert(i + 1, title_node)
                    break
            else: assert False # pragma: no cover

        return [figure_node]

class CodeFigure(_Figure):
    style_classes = ['m-code-figure']

class ConsoleFigure(CodeFigure):
    style_classes = ['m-console-figure']

class MathFigure(_Figure):
    style_classes = ['m-figure']

class GraphFigure(_Figure):
    style_classes = ['m-figure']

class Text(rst.Directive):
    has_content = True
    option_spec = {'class': directives.class_option}

    style_class = ''

    def run(self):
        set_classes(self.options)

        text = '\n'.join(self.content)
        container_node = nodes.container(text, **self.options)
        container_node['classes'] += ['m-text', self.style_class]

        self.state.nested_parse(self.content, self.content_offset,
                                container_node)
        return [container_node]

class DefaultText(Text):
    style_class = 'm-default'

class PrimaryText(Text):
    style_class = 'm-primary'

class SuccessText(Text):
    style_class = 'm-success'

class WarningText(Text):
    style_class = 'm-warning'

class DangerText(Text):
    style_class = 'm-danger'

class InfoText(Text):
    style_class = 'm-info'

class DimText(Text):
    style_class = 'm-dim'

class Button(rst.Directive):
    has_content = True
    final_argument_whitespace = True
    required_arguments = 1
    option_spec = {'class': directives.class_option}

    style_class = ''

    def run(self):
        set_classes(self.options)

        text = '\n'.join(self.content)
        ref_node = nodes.reference(text, '', refuri=self.arguments[0])

        if self.content:
            node = nodes.Element()          # anonymous container for parsing
            self.state.nested_parse(self.content, self.content_offset, node)
            first_node = node[0]
            if not isinstance(first_node, nodes.paragraph):
                error = self.state_machine.reporter.error(
                      'Button must contain directly a caption and optional description',
                      nodes.literal_block(self.block_text, self.block_text),
                      line=self.lineno)
                return [ref_node, error]

            if len(node) == 1:
                caption = nodes.inline(first_node.rawsource, '',
                                       *first_node.children)
            else:
                caption = nodes.container(first_node.rawsource,
                                          *first_node.children)
                caption['classes'] += ['m-big']

            caption.source = first_node.source
            caption.line = first_node.line
            ref_node += caption

            if len(node) > 1:
                second_node = node[1]
                description = nodes.container(second_node.rawsource,
                                              *second_node.children)
                description['classes'] += ['m-small']
                ref_node += description

        wrapper_node = nodes.container(**self.options)
        wrapper_node['classes'] += ['m-button', self.style_class]
        wrapper_node += ref_node

        return [wrapper_node]

class DefaultButton(Button):
    style_class = 'm-default'

class PrimaryButton(Button):
    style_class = 'm-primary'

class SuccessButton(Button):
    style_class = 'm-success'

class WarningButton(Button):
    style_class = 'm-warning'

class DangerButton(Button):
    style_class = 'm-danger'

class InfoButton(Button):
    style_class = 'm-info'

class DimButton(Button):
    style_class = 'm-dim'

class FlatButton(Button):
    style_class = 'm-flat'

def label(style_classes, name, rawtext, text, lineno, inliner, options, content):
    return [nodes.inline(rawtext, text, classes=['m-label'] + style_classes)], []

def label_default(name, rawtext, text, lineno, inliner, options={}, content=[]):
    return label(['m-default'], name, rawtext, text, lineno, inliner, options, content)

def label_primary(name, rawtext, text, lineno, inliner, options={}, content=[]):
    return label(['m-primary'], name, rawtext, text, lineno, inliner, options, content)

def label_success(name, rawtext, text, lineno, inliner, options={}, content=[]):
    return label(['m-success'], name, rawtext, text, lineno, inliner, options, content)

def label_warning(name, rawtext, text, lineno, inliner, options={}, content=[]):
    return label(['m-warning'], name, rawtext, text, lineno, inliner, options, content)

def label_danger(name, rawtext, text, lineno, inliner, options={}, content=[]):
    return label(['m-danger'], name, rawtext, text, lineno, inliner, options, content)

def label_info(name, rawtext, text, lineno, inliner, options={}, content=[]):
    return label(['m-info'], name, rawtext, text, lineno, inliner, options, content)

def label_dim(name, rawtext, text, lineno, inliner, options={}, content=[]):
    return label(['m-dim'], name, rawtext, text, lineno, inliner, options, content)

def label_flat_default(name, rawtext, text, lineno, inliner, options={}, content=[]):
    return label(['m-flat', 'm-default'], name, rawtext, text, lineno, inliner, options, content)

def label_flat_primary(name, rawtext, text, lineno, inliner, options={}, content=[]):
    return label(['m-flat', 'm-primary'], name, rawtext, text, lineno, inliner, options, content)

def label_flat_success(name, rawtext, text, lineno, inliner, options={}, content=[]):
    return label(['m-flat', 'm-success'], name, rawtext, text, lineno, inliner, options, content)

def label_flat_warning(name, rawtext, text, lineno, inliner, options={}, content=[]):
    return label(['m-flat', 'm-warning'], name, rawtext, text, lineno, inliner, options, content)

def label_flat_danger(name, rawtext, text, lineno, inliner, options={}, content=[]):
    return label('m-flat', ['m-danger'], name, rawtext, text, lineno, inliner, options, content)

def label_flat_info(name, rawtext, text, lineno, inliner, options={}, content=[]):
    return label(['m-flat', 'm-info'], name, rawtext, text, lineno, inliner, options, content)

def label_flat_dim(name, rawtext, text, lineno, inliner, options={}, content=[]):
    return label(['m-flat', 'm-dim'], name, rawtext, text, lineno, inliner, options, content)

def register_mcss(**kwargs):
    rst.directives.register_directive('transition', Transition)

    rst.directives.register_directive('note-default', DefaultNote)
    rst.directives.register_directive('note-primary', PrimaryNote)
    rst.directives.register_directive('note-success', SuccessNote)
    rst.directives.register_directive('note-warning', WarningNote)
    rst.directives.register_directive('note-danger', DangerNote)
    rst.directives.register_directive('note-info', InfoNote)
    rst.directives.register_directive('note-dim', DimNote)

    rst.directives.register_directive('block-default', DefaultBlock)
    rst.directives.register_directive('block-primary', PrimaryBlock)
    rst.directives.register_directive('block-success', SuccessBlock)
    rst.directives.register_directive('block-warning', WarningBlock)
    rst.directives.register_directive('block-danger', DangerBlock)
    rst.directives.register_directive('block-info', InfoBlock)
    rst.directives.register_directive('block-dim', DimBlock)
    rst.directives.register_directive('block-flat', FlatBlock)

    rst.directives.register_directive('frame', Frame)
    rst.directives.register_directive('code-figure', CodeFigure)
    rst.directives.register_directive('console-figure', ConsoleFigure)
    rst.directives.register_directive('math-figure', MathFigure)
    rst.directives.register_directive('graph-figure', GraphFigure)

    rst.directives.register_directive('text-default', DefaultText)
    rst.directives.register_directive('text-primary', PrimaryText)
    rst.directives.register_directive('text-success', SuccessText)
    rst.directives.register_directive('text-warning', WarningText)
    rst.directives.register_directive('text-danger', DangerText)
    rst.directives.register_directive('text-info', InfoText)
    rst.directives.register_directive('text-dim', DimText)

    rst.directives.register_directive('button-default', DefaultButton)
    rst.directives.register_directive('button-primary', PrimaryButton)
    rst.directives.register_directive('button-success', SuccessButton)
    rst.directives.register_directive('button-warning', WarningButton)
    rst.directives.register_directive('button-danger', DangerButton)
    rst.directives.register_directive('button-info', InfoButton)
    rst.directives.register_directive('button-dim', DimButton)
    rst.directives.register_directive('button-flat', FlatButton)

    rst.roles.register_canonical_role('label-default', label_default)
    rst.roles.register_canonical_role('label-primary', label_primary)
    rst.roles.register_canonical_role('label-success', label_success)
    rst.roles.register_canonical_role('label-warning', label_warning)
    rst.roles.register_canonical_role('label-danger', label_danger)
    rst.roles.register_canonical_role('label-info', label_info)
    rst.roles.register_canonical_role('label-dim', label_dim)

    rst.roles.register_canonical_role('label-flat-default', label_flat_default)
    rst.roles.register_canonical_role('label-flat-primary', label_flat_primary)
    rst.roles.register_canonical_role('label-flat-success', label_flat_success)
    rst.roles.register_canonical_role('label-flat-warning', label_flat_warning)
    rst.roles.register_canonical_role('label-flat-danger', label_flat_danger)
    rst.roles.register_canonical_role('label-flat-info', label_flat_info)
    rst.roles.register_canonical_role('label-flat-dim', label_flat_dim)

# Below is only Pelican-specific functionality. If Pelican is not found, these
# do nothing.

register = register_mcss # for Pelican



#!/usr/bin/env python3
# This file is a part of marzer/poxy and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/poxy/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT
"""
Helpers for working with HTML using BeautifulSoup.
"""

from typing import List, Union

import bs4

from .utils import *

# =======================================================================================================================
# BS4 HELPER FUNCTIONS
# =======================================================================================================================


def find_parent(tag, names, cutoff=None):
    if not is_collection(names):
        names = (names,)
    parent = tag.parent
    while parent is not None:
        if cutoff is not None and parent is cutoff:
            return None
        if parent.name in names:
            return parent
        parent = parent.parent
    return parent


def destroy_node(node):
    assert node is not None
    if isinstance(node, bs4.NavigableString):
        node.extract()
    else:
        node.decompose()


def replace_tag(tag, new_tag_str):
    assert tag.parent is not None
    newTags = []
    if new_tag_str:
        doc = bs4.BeautifulSoup(new_tag_str, 'html5lib')
        if len(doc.body.contents) > 0:
            newTags = [f for f in doc.body.contents]
            newTags = [f.extract() for f in newTags]
            prev = tag
            for newTag in newTags:
                prev.insert_after(newTag)
                prev = newTag
    destroy_node(tag)
    return newTags


def shallow_search(starting_tag, names, filter=None):
    if isinstance(starting_tag, bs4.NavigableString):
        return []

    if not is_collection(names):
        names = (names,)

    if starting_tag.name in names:
        if filter is None or filter(starting_tag):
            return [starting_tag]

    results = []
    for tag in starting_tag.children:
        if isinstance(tag, bs4.NavigableString):
            continue
        if tag.name in names:
            if filter is None or filter(tag):
                results.append(tag)
        else:
            results = results + shallow_search(tag, names, filter)
    return results


def string_descendants(starting_tag, filter=None):
    if isinstance(starting_tag, bs4.NavigableString):
        if filter is None or filter(starting_tag):
            return [starting_tag]

    results = []
    for tag in starting_tag.children:
        if isinstance(tag, bs4.NavigableString):
            if filter is None or filter(tag):
                results.append(tag)
        else:
            results = results + string_descendants(tag, filter)
    return results


def add_class(tag, classes):
    assert tag is not None
    appended = False
    if 'class' not in tag.attrs:
        tag['class'] = []
    if not is_collection(classes):
        classes = (classes,)
    for class_ in classes:
        if class_ not in tag['class']:
            tag['class'].append(class_)
            appended = True
    return appended


def remove_class(tag, classes):
    assert tag is not None
    removed = False
    if 'class' in tag.attrs:
        if not is_collection(classes):
            classes = (classes,)
        for class_ in classes:
            if class_ in tag['class']:
                tag['class'].remove(class_)
                removed = True
        if removed and len(tag['class']) == 0:
            del tag['class']
    return removed


def set_class(tag, classes):
    assert tag is not None
    tag['class'] = []
    add_class(tag, classes)


def get_classes(tag) -> List[str]:
    assert tag is not None
    if 'class' not in tag.attrs:
        return []
    classes = tag['class']
    if not is_collection(classes):
        classes = [classes]
    return [str(c) for c in classes]


def has_any_classes(tag, *classes) -> bool:
    assert tag is not None
    if 'class' not in tag.attrs:
        return False
    has = get_classes(tag)
    for cls in classes:
        if cls in has:
            return True
    return False


# =======================================================================================================================
# HTML DOCUMENT
# =======================================================================================================================


class HTMLDocument(object):
    def __init__(self, source: Union[str, Path], logger):
        self.__logger = logger

        assert source is not None
        if isinstance(source, Path):
            with open(str(source), 'r', encoding='utf-8') as f:
                self.__doc = bs4.BeautifulSoup(f, 'html5lib', from_encoding='utf-8')
        else:
            self.__doc = bs4.BeautifulSoup(source, 'html5lib')

        self.html = self.__doc.html
        self.head = self.__doc.head
        self.body = self.__doc.body

        self.article = None
        self.article_content = None
        try:
            self.article = self.__doc.body.main.article
            self.article_content = self.article.div.div.div
        except:
            pass

        self.table_of_contents = None
        self.sections = None
        if self.article_content is not None:
            for toc_tag in ('nav', 'div'):
                for tag in self.article_content(toc_tag, class_='m-block m-default', recursive=False):
                    if tag.h3 and tag.h3.string == 'Contents':
                        self.table_of_contents = tag
                        break
                if self.table_of_contents is not None:
                    break
            self.sections = self.article_content('section', recursive=False)

    def smooth(self):
        self.__doc.smooth()

    def write(self, path):
        log(self.__logger, rf'Writing {path}')
        with open(self.path, 'w', encoding='utf-8', newline='\n') as f:
            f.write(str(self.__doc))

    def __str__(self) -> str:
        return str(self.__doc)

    def new_tag(self, tag_name, parent=None, string=None, class_=None, index=None, before=None, after=None, **kwargs):
        tag = self.__doc.new_tag(tag_name, **kwargs)
        if string is not None:
            if tag.string is not None:
                tag.string.replace_with(string)
            else:
                tag.string = bs4.NavigableString(string)
        if class_ is not None:
            tag['class'] = class_
        if before is not None:
            before.insert_before(tag)
        elif after is not None:
            after.insert_after(tag)
        elif parent is not None:
            if index is None or index < 0:
                parent.append(tag)
            else:
                parent.insert(index, tag)

        return tag

    def find_all_from_sections(self, name=None, select=None, section=None, include_toc=False, **kwargs):
        tags = []
        if self.article_content is not None:
            sections = None
            if section is not None:
                sections = self.article_content('section', recursive=False, id='section')
            else:
                sections = self.sections
            if include_toc and self.table_of_contents is not None:
                sections = [self.table_of_contents, *sections]
            for sect in sections:
                matches = sect(name, **kwargs) if name is not None else [sect]
                if select is not None:
                    newMatches = []
                    for match in matches:
                        newMatches += match.select(select)
                    matches = newMatches
                tags += matches
        return tags



#
#   This file is part of m.css.
#
#   Copyright © 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024
#             Vladimír Vondruš <mosra@centrum.cz>
#
#   Permission is hereby granted, free of charge, to any person obtaining a
#   copy of this software and associated documentation files (the "Software"),
#   to deal in the Software without restriction, including without limitation
#   the rights to use, copy, modify, merge, publish, distribute, sublicense,
#   and/or sell copies of the Software, and to permit persons to whom the
#   Software is furnished to do so, subject to the following conditions:
#
#   The above copyright notice and this permission notice shall be included
#   in all copies or substantial portions of the Software.
#
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#   DEALINGS IN THE SOFTWARE.
#

import os
import gzip
from docutils import nodes
from docutils.parsers import rst
from docutils.parsers.rst.roles import set_classes

settings = {}

def filesize(name, rawtext, text, lineno, inliner, options={}, content=[]):
    # Support both {filename} (3.7.1) and {static} (3.8) placeholders
    size = os.path.getsize(text.format(filename=settings['INPUT'], static=settings['INPUT']))

    for unit in ['','k','M','G','T']:
        if abs(size) < 1024.0:
            size_string = "%3.1f %sB" % (size, unit)
            break
        size /= 1024.0
    else: size_string = "%.1f PB" % size

    set_classes(options)
    return [nodes.inline(size_string, size_string, **options)], []

def filesize_gz(name, rawtext, text, lineno, inliner, options={}, content=[]):
    # Support both {filename} (3.7.1) and {static} (3.8) placeholders
    with open(text.format(filename=settings['INPUT'], static=settings['INPUT']), mode='rb') as f:
        size = len(gzip.compress(f.read()))

    for unit in ['','k','M','G','T']:
        if abs(size) < 1024.0:
            size_string = "%3.1f %sB" % (size, unit)
            break
        size /= 1024.0
    else: size_string = "%.1f PB" % size

    set_classes(options)
    return [nodes.inline(size_string, size_string, **options)], []

def register_mcss(mcss_settings, **kwargs):
    global settings
    settings['INPUT'] = mcss_settings['INPUT']
    rst.roles.register_local_role('filesize', filesize)
    rst.roles.register_local_role('filesize-gz', filesize_gz)

# Below is only Pelican-specific functionality. If Pelican is not found, these
# do nothing.

def _pelican_configure(pelicanobj):
    settings = {
        'INPUT': os.path.join(os.getcwd(), pelicanobj.settings['PATH'])
    }
    register_mcss(mcss_settings=settings)

def register(): # for Pelican
    from pelican import signals

    signals.initialized.connect(_pelican_configure)



#
#   This file is part of m.css.
#
#   Copyright © 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024
#             Vladimír Vondruš <mosra@centrum.cz>
#
#   Permission is hereby granted, free of charge, to any person obtaining a
#   copy of this software and associated documentation files (the "Software"),
#   to deal in the Software without restriction, including without limitation
#   the rights to use, copy, modify, merge, publish, distribute, sublicense,
#   and/or sell copies of the Software, and to permit persons to whom the
#   Software is furnished to do so, subject to the following conditions:
#
#   The above copyright notice and this permission notice shall be included
#   in all copies or substantial portions of the Software.
#
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#   DEALINGS IN THE SOFTWARE.
#

import copy
import logging
import os.path
import re

from docutils.writers.html5_polyglot import HTMLTranslator
from docutils.transforms import Transform
import docutils
from docutils import nodes, utils
from docutils.utils import smartquotes
from urllib.parse import urljoin

try:
    import pyphen
except ImportError:
    pyphen = None

logger = logging.getLogger(__name__)

default_settings = {
    # Language used for hyphenation and smart quotes
    'M_HTMLSANITY_LANGUAGE': 'en',
    # Extra docutils settings
    'M_HTMLSANITY_DOCUTILS_SETTINGS': {},
    # Whether hyphenation is enabled
    'M_HTMLSANITY_HYPHENATION': False,
    # Whether smart quotes are enabled
    'M_HTMLSANITY_SMART_QUOTES': False,
    # List of formatted document-level fields (which should get hyphenation
    # and smart quotes applied as well)
    'M_HTMLSANITY_FORMATTED_FIELDS': []
}

settings = None

# Default docutils settings. Some things added in register() / register_mcss(),
# and from the M_HTMLSANITY_DOCUTILS_SETTINGS option.
docutils_settings = {
    'initial_header_level': '2',
    'syntax_highlight': 'short',
    'input_encoding': 'utf-8',
    'halt_level': 2,
    'traceback': True,
    'embed_stylesheet': False
}

# findall() is new in docutils 0.18, replacing traverse() because that one was
# attempted to be changed to return an iterator instead of a list, which broke
# Sphinx, so it instead got deprecated in favor of findall() which is the same
# as traverse() returning an iterator was. To retain compatibility with earlier
# versions, add an alias.
if not hasattr(docutils.nodes.Node, 'findall'):
    setattr(docutils.nodes.Node, 'findall', docutils.nodes.Node.traverse)

words_re = re.compile(r'\w+', re.UNICODE|re.X)

def extract_document_language(document):
    # Take the one from settings as default
    language = document.settings.language_code

    # Then try to find the :lang: metadata option
    for field in document.findall(nodes.field):
        assert isinstance(field[0], nodes.field_name)
        assert isinstance(field[1], nodes.field_body)
        # field_body -> paragraph -> text
        if field[0][0] == 'lang': return str(field[1][0][0])

    return language

def can_apply_typography(txtnode):
    # Exclude:
    #  - literals and spans inside literals
    #  - raw code (such as SVG)
    #  - field names
    #  - bibliographic elements (author, date, ... fields)
    #  - links with title that's the same as URL (or e-mail)
    if isinstance(txtnode.parent, nodes.literal) or \
       isinstance(txtnode.parent.parent, nodes.literal) or \
       isinstance(txtnode.parent, nodes.raw) or \
       isinstance(txtnode.parent, nodes.field_name) or \
       isinstance(txtnode.parent, nodes.Bibliographic) or \
       (isinstance(txtnode.parent, nodes.reference) and
            (txtnode.astext() == txtnode.parent.get('refuri', '') or 'mailto:' + txtnode.astext() == txtnode.parent.get('refuri', ''))):
        return False

    # From fields include only the ones that are in M_HTMLSANITY_FORMATTED_FIELDS
    if isinstance(txtnode.parent.parent, nodes.field_body):
        field_name_index = txtnode.parent.parent.parent.first_child_matching_class(nodes.field_name)
        if txtnode.parent.parent.parent[field_name_index][0] in settings['M_HTMLSANITY_FORMATTED_FIELDS']:
            return True
        return False

    return True

class SmartQuotes(docutils.transforms.universal.SmartQuotes):
    """Smart quote transform

    Copy-paste of docutils builtin smart_quotes feature, patched to *not*
    replace quotes etc. in inline code blocks.  Original source from
    https://github.com/docutils-mirror/docutils/blob/e88c5fb08d5cdfa8b4ac1020dd6f7177778d5990/docutils/transforms/universal.py#L207

    Contrary to the builtin, this is controlled via HTMLSANITY_SMART_QUOTES
    instead of smart_quotes in DOCUTILS_SETTINGS, so the original smart quotes
    implementation is not executed.
    """

    def apply(self):
        # We are using our own config variable instead of
        # self.document.settings.smart_quotes in order to avoid the builtin
        # SmartQuotes to be executed as well
        if not settings['M_HTMLSANITY_SMART_QUOTES']:
            return
        try:
            alternative = settings['M_HTMLSANITY_SMART_QUOTES'].startswith('alt')
        except AttributeError:
            alternative = False
        # print repr(alternative)

        document_language = extract_document_language(self.document)

        # "Educate" quotes in normal text. Handle each block of text
        # (TextElement node) as a unit to keep context around inline nodes:
        for node in self.document.findall(nodes.TextElement):
            # skip preformatted text blocks and special elements:
            if isinstance(node, (nodes.FixedTextElement, nodes.Special)):
                continue
            # nested TextElements are not "block-level" elements:
            if isinstance(node.parent, nodes.TextElement):
                continue

            # list of text nodes in the "text block":
            # Patched here to exclude more stuff.
            txtnodes = []
            for txtnode in node.findall(nodes.Text):
                if not can_apply_typography(txtnode): continue
                # Don't convert -- in option strings
                if isinstance(txtnode.parent, nodes.option_string): continue

                txtnodes += [txtnode]

            # language: use typographical quotes for language "lang"
            lang = node.get_language_code(document_language)
            # use alternative form if `smart-quotes` setting starts with "alt":
            if alternative:
                if '-x-altquot' in lang:
                    lang = lang.replace('-x-altquot', '')
                else:
                    lang += '-x-altquot'
            # drop subtags missing in quotes:
            for tag in utils.normalize_language_tag(lang):
                if tag in smartquotes.smartchars.quotes:
                    lang = tag
                    break
            else: # language not supported: (keep ASCII quotes)
                lang = ''

            # Iterator educating quotes in plain text:
            # '2': set all, using old school en- and em- dash shortcuts
            teacher = smartquotes.educate_tokens(self.get_tokens(txtnodes),
                                                 attr='2', language=lang)

            for txtnode, newtext in zip(txtnodes, teacher):
                txtnode.parent.replace(txtnode, nodes.Text(newtext))

            self.unsupported_languages = set() # reset

class Pyphen(Transform):
    """Hyphenation using pyphen

    Enabled or disabled using HTMLSANITY_HYPHENATION boolean option, defaulting
    to ``False``.
    """

    # Max Docutils priority is 990, be sure that this is applied at the very
    # last
    default_priority = 991

    def __init__(self, document, startnode):
        Transform.__init__(self, document, startnode=startnode)

    def apply(self):
        if not settings['M_HTMLSANITY_HYPHENATION']:
            return

        document_language = extract_document_language(self.document)

        pyphen_for_lang = {}

        # Go through all text words and hyphenate them
        for node in self.document.findall(nodes.TextElement):
            # Skip preformatted text blocks and special elements
            if isinstance(node, (nodes.FixedTextElement, nodes.Special)):
                continue

            for txtnode in node.findall(nodes.Text):
                if not can_apply_typography(txtnode): continue

                # Don't hyphenate document title. Not part of
                # can_apply_typography() because we *do* want smart quotes for
                # a document title.
                if isinstance(txtnode.parent, nodes.title): continue

                # Useful for debugging, don't remove ;)
                #print(repr(txtnode.parent), repr(txtnode.parent.parent), repr(txtnode.parent.parent.parent))

                # Proper language-dependent hyphenation. Can't be done for
                # `node` as a paragraph can consist of more than one language.
                lang = txtnode.parent.get_language_code(document_language)

                # Create new Pyphen object for given lang, if not yet cached.
                # I'm assuming this is faster than recreating the instance for
                # every text node
                if lang not in pyphen_for_lang:
                    if not pyphen or lang not in pyphen.LANGUAGES: continue
                    pyphen_for_lang[lang] = pyphen.Pyphen(lang=lang)

                txtnode.parent.replace(txtnode, nodes.Text(words_re.sub(lambda m: pyphen_for_lang[lang].inserted(m.group(0), '\u00AD'), txtnode.astext())))

class SaneHtmlTranslator(HTMLTranslator):
    """Sane HTML translator

    Patched version of docutils builtin HTML5 translator, improving the output
    further.
    """

    # Overrides the ones from docutils HTML5 writer to enable the soft hyphen
    # character
    special_characters = {ord('&'): '&amp;',
                          ord('<'): '&lt;',
                          ord('"'): '&quot;',
                          ord('>'): '&gt;',
                          ord('@'): '&#64;', # may thwart address harvesters
                          ord('\u00AD'): '&shy;'}

    def __init__(self, document):
        HTMLTranslator.__init__(self, document)

        # There's a minor difference between docutils 0.13 and 0.14 that breaks
        # stuff. Monkey-patch it here.
        if not hasattr(self, 'in_word_wrap_point'):
            self.in_word_wrap_point = HTMLTranslator.sollbruchstelle

        # Used by depart_caption() and depart_figure(), see there for details
        self.in_figure_caption_with_description = False

    # Somehow this does the trick and removes docinfo from the body. Was
    # present in the HTML4 translator but not in the HTML5 one, so copying it
    # verbatim over
    def visit_docinfo(self, node):
        self.context.append(len(self.body))
        self.body.append(self.starttag(node, 'table',
                                       CLASS='docinfo',
                                       frame="void", rules="none"))
        self.body.append('<col class="docinfo-name" />\n'
                         '<col class="docinfo-content" />\n'
                         '<tbody valign="top">\n')
        self.in_docinfo = True

    def depart_docinfo(self, node):
        self.body.append('</tbody>\n</table>\n')
        self.in_docinfo = False
        start = self.context.pop()
        self.docinfo = self.body[start:]
        self.body = []

    # Have <abbr> properly with title
    def visit_abbreviation(self, node):
        attrs = {}
        if node.hasattr('title'):
            attrs['title'] = node['title']
        self.body.append(self.starttag(node, 'abbr', '', **attrs))

    def depart_abbreviation(self, node):
        self.body.append('</abbr>')

    # Convert outdated width/height attributes to CSS styles, handle the scale
    # directly inside m.images; don't put URI in alt text.
    def visit_image(self, node):
        atts = {}
        uri = node['uri']
        ext = os.path.splitext(uri)[1].lower()
        atts['src'] = uri
        if 'alt' in node: atts['alt'] = node['alt']
        style = []
        if node.get('width'):
            style += ['width: {}'.format(node['width'])]
        if node.get('height'):
            style += ['height: {}'.format(node['height'])]
        if style:
            atts['style'] = '; '.join(style)
        if (isinstance(node.parent, nodes.TextElement) or
            (isinstance(node.parent, nodes.reference) and
             not isinstance(node.parent.parent, nodes.TextElement))):
            # Inline context or surrounded by <a>...</a>.
            suffix = ''
        else:
            suffix = '\n'
        self.body.append(self.emptytag(node, 'img', suffix, **atts))

    def depart_image(self, node):
        pass

    # Use HTML5 <section> tag for sections (instead of <div class="section">)
    def visit_section(self, node):
        self.section_level += 1
        self.body.append(
            self.starttag(node, 'section'))

    def depart_section(self, node):
        self.section_level -= 1
        self.body.append('</section>\n')

    # Legend inside figure. This is handled in a special way inside
    # depart_caption() and depart_figure() already, no need to add any extra
    # tag again.
    def visit_legend(self, node):
        pass

    def depart_legend(self, node):
        pass

    # Literal -- print as <code> (instead of some <span>)
    def visit_literal(self, node):
        self.body.append(self.starttag(node, 'code', ''))

    def depart_literal(self, node):
        self.body.append('</code>')

    # Literal block -- use <pre> without nested <code> and useless classes
    def visit_literal_block(self, node):
        self.body.append(self.starttag(node, 'pre', ''))

    def depart_literal_block(self, node):
        self.body.append('</pre>\n')

    # Anchor -- drop the useless classes
    def visit_reference(self, node):
        atts = {}
        if 'refuri' in node:
            atts['href'] = node['refuri']
            if ( self.settings.cloak_email_addresses
                 and atts['href'].startswith('mailto:')):
                atts['href'] = self.cloak_mailto(atts['href'])
                self.in_mailto = True
        else:
            assert 'refid' in node, \
                   'References must have "refuri" or "refid" attribute.'
            atts['href'] = '#' + node['refid']
        # why?!?!
        #if not isinstance(node.parent, nodes.TextElement):
            #assert len(node) == 1 and isinstance(node[0], nodes.image)

        # If the link is a plain URL without explicitly specified title, apply
        # m-link-wrap so it doesn't leak out of the view on narrow screens.
        # This can be disabled by explicitly providing the URL also as a title
        # --- then the node will have a name attribute and we'll skip in that
        # case.
        if len(node.children) == 1 and isinstance(node.children[0], nodes.Text) and 'name' not in node and 'refuri' in node and node.children[0] == node['refuri']:
            node['classes'] += ['m-link-wrap']

        self.body.append(self.starttag(node, 'a', '', **atts))

    def depart_reference(self, node):
        self.body.append('</a>')
        if not isinstance(node.parent, nodes.TextElement):
            self.body.append('\n')
        self.in_mailto = False

    # Use <aside> for general topics, <nav> for table of contents (but drop the
    # contents class and ID)
    def visit_topic(self, node):
        self.topic_classes = node['classes']

        if 'contents' in node['classes']:
            node.html_tagname = 'nav'
            node['classes'].remove('contents')
            # If the TOC has a title, the ID will be different, and in that
            # case we'll leave it there.
            if 'contents' in node['ids']: node['ids'].remove('contents')
        else:
            node.html_tagname = 'aside'

        self.body.append(self.starttag(node, node.html_tagname))

    def depart_topic(self, node):
        self.body.append('</{}>\n'.format(node.html_tagname))
        self.topic_classes = []

    # Don't use <colgroup> or other shit in tables
    def visit_colspec(self, node):
        self.colspecs.append(node)
        # "stubs" list is an attribute of the tgroup element:
        node.parent.stubs.append(node.attributes.get('stub'))

    def depart_colspec(self, node):
        # write out <colgroup> when all colspecs are processed
        if isinstance(node.next_node(descend=False, siblings=True),
                      nodes.colspec):
            return
        if 'colwidths-auto' in node.parent.parent['classes'] or (
            'colwidths-auto' in self.settings.table_style and
            ('colwidths-given' not in node.parent.parent['classes'])):
            return

    # Verbatim copied, removing the 'head' class from <th>
    def visit_entry(self, node):
        atts = {'class': []}
        if node.parent.parent.parent.stubs[node.parent.column]:
            # "stubs" list is an attribute of the tgroup element
            atts['class'].append('stub')
        if atts['class'] or isinstance(node.parent.parent, nodes.thead):
            tagname = 'th'
            atts['class'] = ' '.join(atts['class'])
        else:
            tagname = 'td'
            del atts['class']
        node.parent.column += 1
        if 'morerows' in node:
            atts['rowspan'] = node['morerows'] + 1
        if 'morecols' in node:
            atts['colspan'] = node['morecols'] + 1
            node.parent.column += node['morecols']
        self.body.append(self.starttag(node, tagname, '', **atts))
        self.context.append('</%s>\n' % tagname.lower())

    # Don't put comments into the HTML output
    def visit_comment(self, node,
                      sub=re.compile('-(?=-)').sub):
        raise nodes.SkipNode

    # Containers don't need those stupid "docutils" class names
    def visit_container(self, node):
        atts = {}
        self.body.append(self.starttag(node, 'div', **atts))

    def depart_container(self, node):
        self.body.append('</div>\n')

    # Use HTML5 <figure> element for figures
    def visit_figure(self, node):
        atts = {}
        if node.get('id'):
            atts['ids'] = [node['id']]
        style = []
        if node.get('width'):
            style += ['width: {}'.format(node['width'])]
        if style:
            atts['style'] = '; '.join(style)
        self.body.append(self.starttag(node, 'figure', **atts))

    def depart_figure(self, node):
        # See depart_caption() below for details
        if self.in_figure_caption_with_description:
            self.body.append('</div>\n</figcaption>\n')
            self.in_figure_caption_with_description = False
        self.body.append('</figure>\n')

    def visit_caption(self, node):
        self.body.append(self.starttag(node, 'figcaption', ''))

    def depart_caption(self, node):
        # If this is a .m-figure and there's more content after a <figcaption>,
        # we have to put all that into <figcaption> as well, otherwise FF will
        # ignore it (due to `display: table-caption` being set for all
        # .m-figure children, which apparently makes FF render just the first
        # element). To avoid all that content styled as a caption, put it
        # inside a .m-figure-description that undoes the styling for
        # <figcaption>.
        # TODO this may have false positives if there are reST comments and
        # such, figure out a way to query if there are useful nodes. Can't
        # check for just nodes.legend, as there can be arbitrary other stuff.
        if 'classes' in node.parent and 'm-figure' in node.parent['classes'] and node.next_node(descend=False, siblings=True) is not None:
            self.body.append(self.starttag(node, 'div', CLASS='m-figure-description'))
            self.in_figure_caption_with_description = True
        else:
            self.body.append('</figcaption>\n')

    # Line blocks are <p> with lines separated using simple <br />. No need for
    # nested <div>s.
    def visit_line(self, node):
        pass

    def depart_line(self, node):
        self.body.append('<br />\n')

    # Footnote list. Replacing the classes with just .m-footnote.
    def visit_footnote(self, node):
        previous_node = node.parent[node.parent.index(node)-1]
        if not isinstance(previous_node, type(node)):
            self.body.append('<dl class="m-footnote">\n')

    def depart_footnote(self, node):
        self.body.append('</dd>\n')
        if not isinstance(node.next_node(descend=False, siblings=True),
                          type(node)):
            self.body.append('</dl>\n')

    # Footnote reference
    def visit_footnote_reference(self, node):
        href = '#' + node['refid']
        self.body.append(self.starttag(node, 'a', '', CLASS='m-footnote', href=href))

    def depart_footnote_reference(self, node):
        self.body.append('</a>')

    # Footnote and citation labels
    def visit_label(self, node):
        self.body.append(self.starttag(node.parent, 'dt', ''))

    def depart_label(self, node):
        if self.settings.footnote_backlinks:
            backrefs = node.parent['backrefs']
            if len(backrefs) == 1:
                self.body.append('</a>')
        self.body.append('.</dt>\n<dd><span class="m-footnote">')
        if len(backrefs) == 1:
            self.body.append('<a href="#{}">^</a>'.format(backrefs[0]))
        else:
            self.body.append('^ ')
            self.body.append(format(' '.join('<a href="#{}">{}</a>'.format(ref, chr(ord('a') + i)) for i, ref in enumerate(backrefs))))
        self.body.append('</span> ')

    def visit_line_block(self, node):
        self.body.append(self.starttag(node, 'p'))

    def depart_line_block(self, node):
        self.body.append('</p>\n')

    # Copied from the HTML4 translator, somehow not present in the HTML5 one.
    # Not having this generates *a lot* of <p> tags everywhere.
    def should_be_compact_paragraph(self, node):
        """
        Determine if the <p> tags around paragraph ``node`` can be omitted.
        """
        if (isinstance(node.parent, nodes.document) or
            isinstance(node.parent, nodes.compound)):
            # Never compact paragraphs in document or compound
            return False
        for key, value in node.attlist():
            if (node.is_not_default(key) and
                not (key == 'classes' and value in
                     ([], ['first'], ['last'], ['first', 'last']))):
                # Attribute which needs to survive.
                return False
        first = isinstance(node.parent[0], nodes.label) # skip label
        for child in node.parent.children[first:]:
            # only first paragraph can be compact
            if isinstance(child, nodes.Invisible):
                continue
            if child is node:
                break
            return False
        parent_length = len([n for n in node.parent if not isinstance(
            n, (nodes.Invisible, nodes.label))])
        if ( self.compact_simple
             or self.compact_field_list
             or self.compact_p and parent_length == 1):
            return True
        return False

    def visit_paragraph(self, node):
        if self.should_be_compact_paragraph(node):
            self.context.append('')
        else:
            self.body.append(self.starttag(node, 'p', ''))
            self.context.append('</p>\n')

    def depart_paragraph(self, node):
        self.body.append(self.context.pop())

    # Titles in topics should be <h3>
    def visit_title(self, node):
        """Only 6 section levels are supported by HTML."""
        check_id = 0  # TODO: is this a bool (False) or a counter?
        close_tag = '</p>\n'
        if isinstance(node.parent, nodes.topic):
            self.body.append(
                  self.starttag(node, 'h3', ''))
            close_tag = '</h3>\n'
        elif isinstance(node.parent, nodes.sidebar):
            self.body.append(
                  self.starttag(node, 'p', '', CLASS='sidebar-title'))
        elif isinstance(node.parent, nodes.Admonition):
            self.body.append(
                  self.starttag(node, 'p', '', CLASS='admonition-title'))
        elif isinstance(node.parent, nodes.table):
            self.body.append(
                  self.starttag(node, 'caption', ''))
            close_tag = '</caption>\n'
        elif isinstance(node.parent, nodes.document):
            self.body.append(self.starttag(node, 'h1', '', CLASS='title'))
            close_tag = '</h1>\n'
            self.in_document_title = len(self.body)
        else:
            assert isinstance(node.parent, nodes.section)
            h_level = self.section_level + self.initial_header_level - 1
            atts = {}
            if (len(node.parent) >= 2 and
                isinstance(node.parent[1], nodes.subtitle)):
                atts['CLASS'] = 'with-subtitle'
            self.body.append(
                  self.starttag(node, 'h%s' % h_level, '', **atts))
            atts = {}
            if node.hasattr('refid'):
                atts['href'] = '#' + node['refid']
            if atts:
                self.body.append(self.starttag({}, 'a', '', **atts))
                close_tag = '</a></h%s>\n' % (h_level)
            else:
                close_tag = '</h%s>\n' % (h_level)
        self.context.append(close_tag)

    def depart_title(self, node):
        self.body.append(self.context.pop())
        if self.in_document_title:
            self.title = self.body[self.in_document_title:-1]
            self.in_document_title = 0
            self.body_pre_docinfo.extend(self.body)
            self.html_title.extend(self.body)
            del self.body[:]

    # <ul>, <ol>, <dl> -- verbatim copied, removing "simple" class. For <ol>
    # also removing the enumtype
    def visit_bullet_list(self, node):
        atts = {}
        old_compact_simple = self.compact_simple
        self.context.append((self.compact_simple, self.compact_p))
        self.compact_p = None
        self.compact_simple = self.is_compactable(node)
        self.body.append(self.starttag(node, 'ul', **atts))

    def depart_bullet_list(self, node):
        self.compact_simple, self.compact_p = self.context.pop()
        self.body.append('</ul>\n')

    def visit_enumerated_list(self, node):
        atts = {}
        if 'start' in node:
            atts['start'] = node['start']
        self.body.append(self.starttag(node, 'ol', **atts))

    def depart_enumerated_list(self, node):
        self.body.append('</ol>\n')

    def visit_definition_list(self, node):
        classes = node.setdefault('classes', [])
        self.body.append(self.starttag(node, 'dl'))

    def depart_definition_list(self, node):
        self.body.append('</dl>\n')

    # no class="docutils" in <hr>; if it's m.css .. transition:: directive
    # (having children), put it in a <p> instead
    def visit_transition(self, node):
        if len(node.children) > 0:
            self.body.append(self.starttag(node, 'p', ''))
        else:
            self.body.append(self.emptytag(node, 'hr'))

    def depart_transition(self, node):
        if len(node.children) > 0:
            self.body.append('</p>\n')

class _SaneFieldBodyTranslator(SaneHtmlTranslator):
    """
    Copy of _FieldBodyTranslator with the only difference being inherited from
    SaneHtmlTranslator instead of HTMLTranslator
    """

    def __init__(self, document):
        SaneHtmlTranslator.__init__(self, document)

    # Overriding the function in SaneHtmlTranslator, in addition never
    # compacting paragraphs directly in field bodies (such as article summary
    # or page footer) unless explicitly told it so. The sad thing is that the
    # Pelican theme currently always expects the summaries to be wrapped in
    # <p>, while the Python docs expect exactly the other case.
    def should_be_compact_paragraph(self, node):
        if isinstance(node.parent, nodes.field_body) and not self.compact_field_list:
            return False
        return SaneHtmlTranslator.should_be_compact_paragraph(self, node)

    def astext(self):
        return ''.join(self.body)

    # If this wouldn't be here, the output would have <dd> around. Not useful.
    def visit_field_body(self, node):
        pass

    def depart_field_body(self, node):
        pass

class SaneHtmlWriter(docutils.writers.html5_polyglot.Writer):
    def __init__(self):
        docutils.writers.html5_polyglot.Writer.__init__(self)

        self.translator_class = SaneHtmlTranslator

    def get_transforms(self):
        return docutils.writers.html5_polyglot.Writer.get_transforms(self) + [SmartQuotes, Pyphen]

def render_rst(value):
    pub = docutils.core.Publisher(
        writer=SaneHtmlWriter(),
        source_class=docutils.io.StringInput,
        destination_class=docutils.io.StringOutput)
    pub.set_components('standalone', 'restructuredtext', 'html')
    pub.writer.translator_class = _SaneFieldBodyTranslator
    pub.process_programmatic_settings(None, docutils_settings, None)
    pub.set_source(source=value)
    pub.publish(enable_exit_status=True)
    return pub.writer.parts.get('body').strip()

def rtrim(value):
    return value.rstrip()

def hyphenate(value, enable=None, lang=None):
    if enable is None: enable = settings['M_HTMLSANITY_HYPHENATION']
    if lang is None: lang = settings['M_HTMLSANITY_LANGUAGE']
    if not enable or not pyphen: return value
    pyphen_ = pyphen.Pyphen(lang=lang)
    return words_re.sub(lambda m: pyphen_.inserted(m.group(0), '&shy;'), str(value))

def dehyphenate(value, enable=None):
    if enable is None: enable = settings['M_HTMLSANITY_HYPHENATION']
    if not enable: return value
    return value.replace('&shy;', '')

def register_mcss(mcss_settings, jinja_environment, **kwargs):
    global default_settings, settings
    settings = copy.deepcopy(default_settings)
    for key in settings.keys():
        if key in mcss_settings: settings[key] = mcss_settings[key]
    docutils_settings['language_code'] = settings['M_HTMLSANITY_LANGUAGE']
    docutils_settings.update(settings['M_HTMLSANITY_DOCUTILS_SETTINGS'])

    jinja_environment.filters['rtrim'] = rtrim
    jinja_environment.filters['render_rst'] = render_rst
    jinja_environment.filters['hyphenate'] = hyphenate
    jinja_environment.filters['dehyphenate'] = dehyphenate

# Below is only Pelican-specific functionality. If Pelican is not found, these
# do nothing.
try:
    from pelican import signals
    from pelican.readers import RstReader

    class PelicanSaneRstReader(RstReader):
        writer_class = SaneHtmlWriter
        field_body_translator_class = _SaneFieldBodyTranslator

except ImportError:
    pass

pelican_settings = {}

def pelican_expand_link(link, content):
    link_regex = r"""^
        (?P<markup>)(?P<quote>)
        (?P<path>{0}(?P<value>.*))
        $""".format(pelican_settings['INTRASITE_LINK_REGEX'])
    links = re.compile(link_regex, re.X)
    return links.sub(
        lambda m: content._link_replacer(content.get_siteurl(), m),
        link)

def pelican_expand_links(text, content):
    # TODO: fields that are in FORMATTED_FIELDS are already expanded, but that
    # requires extra work on user side. Ideal would be to handle that all on
    # template side, so keeping this one for the time when we can replace
    # FORMATTED_FIELDS with render_rst as well.
    return content._update_content(text, content.get_siteurl())

# To be consistent with both what Pelican does now with '/'.join(SITEURL, url)
# and with https://github.com/getpelican/pelican/pull/2196. Keep consistent
# with m.alias.
def pelican_format_siteurl(url):
    return urljoin(pelican_settings['SITEURL'] + ('/' if not pelican_settings['SITEURL'].endswith('/') else ''), url)

def _pelican_configure(pelicanobj):
    pelicanobj.settings['JINJA_FILTERS']['rtrim'] = rtrim
    pelicanobj.settings['JINJA_FILTERS']['render_rst'] = render_rst
    pelicanobj.settings['JINJA_FILTERS']['expand_link'] = pelican_expand_link
    pelicanobj.settings['JINJA_FILTERS']['expand_links'] = pelican_expand_links
    pelicanobj.settings['JINJA_FILTERS']['format_siteurl'] = pelican_format_siteurl
    pelicanobj.settings['JINJA_FILTERS']['hyphenate'] = hyphenate
    pelicanobj.settings['JINJA_FILTERS']['dehyphenate'] = dehyphenate

    # Map the setting key names from Pelican's own
    global default_settings, settings
    settings = copy.deepcopy(default_settings)
    for key, pelicankey in [
        ('M_HTMLSANITY_LANGUAGE', 'DEFAULT_LANG'),
        ('M_HTMLSANITY_DOCUTILS_SETTINGS', 'DOCUTILS_SETTINGS'),
        ('M_HTMLSANITY_FORMATTED_FIELDS', 'FORMATTED_FIELDS')]:
        settings[key] = pelicanobj.settings[pelicankey]

    # Settings with the same name both here and in Pelican
    for key in 'M_HTMLSANITY_HYPHENATION', 'M_HTMLSANITY_SMART_QUOTES':
        if key in pelicanobj.settings: settings[key] = pelicanobj.settings[key]

    # Save settings needed only for Pelican-specific functionality
    global pelican_settings
    for key in 'INTRASITE_LINK_REGEX', 'SITEURL':
        pelican_settings[key] = pelicanobj.settings[key]

    # Update the docutils settings using the above
    docutils_settings['language_code'] = settings['M_HTMLSANITY_LANGUAGE']
    docutils_settings.update(settings['M_HTMLSANITY_DOCUTILS_SETTINGS'])

def _pelican_add_reader(readers):
    readers.reader_classes['rst'] = PelicanSaneRstReader

def register(): # for Pelican
    signals.initialized.connect(_pelican_configure)
    signals.readers_init.connect(_pelican_add_reader)



#
#   This file is part of m.css.
#
#   Copyright © 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024
#             Vladimír Vondruš <mosra@centrum.cz>
#
#   Permission is hereby granted, free of charge, to any person obtaining a
#   copy of this software and associated documentation files (the "Software"),
#   to deal in the Software without restriction, including without limitation
#   the rights to use, copy, modify, merge, publish, distribute, sublicense,
#   and/or sell copies of the Software, and to permit persons to whom the
#   Software is furnished to do so, subject to the following conditions:
#
#   The above copyright notice and this permission notice shall be included
#   in all copies or substantial portions of the Software.
#
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#   DEALINGS IN THE SOFTWARE.
#

import logging
import os

from pelican import signals

logger = logging.getLogger(__name__)

_names = {'authors': "Author",
          'categories': "Category",
          'tags': "Tag"}

def _read_pages(article_generator, key):
    path = article_generator.settings.get(*key)
    fullpath = os.path.join(article_generator.settings['PATH'], path)
    if not os.path.isdir(fullpath): return {}

    pages = {}
    for f in os.listdir(fullpath):
        fullf = os.path.join(fullpath, f)
        if not os.path.isfile(fullf): continue

        logger.debug("Read file {} -> {}".format(os.path.join(path, f), {
            'authors': "Author",
            'categories': "Category",
            'tags': "Tag"}[key[1]]))

        base_path, filename = os.path.split(fullf)
        pages[os.path.splitext(filename)[0]] = article_generator.readers.read_file(base_path, filename, context=article_generator.context)
    return pages

def populate_metadata(article_generator):
    authors = _read_pages(article_generator, ('M_METADATA_AUTHOR_PATH', 'authors'))
    categories = _read_pages(article_generator, ('M_METADATA_CATEGORY_PATH', 'categories'))
    tags = _read_pages(article_generator, ('M_METADATA_TAG_PATH', 'tags'))

    for author, _ in article_generator.authors:
        author.page = authors.get(author.slug, {})
    for category, _ in article_generator.categories:
        category.page = categories.get(category.slug, {})
    for tag in article_generator.tags:
        tag.page = tags.get(tag.slug, {})

    for article in article_generator.articles:
        page = authors.get(article.author.slug, {})
        for i in ['badge', 'image', 'twitter', 'twitter_id']:
            if hasattr(page, i): setattr(article.author, i, getattr(page, i))
        if hasattr(page, 'title'):
            article.author.badge_title = page.title

        page = categories.get(article.category.slug, {})
        for i in ['badge', 'image']:
            if hasattr(page, i): setattr(article.category, i, getattr(page, i))
        if hasattr(page, 'title'):
            article.category.badge_title = page.title

def register_mcss(**kwargs):
    assert not "This plugin is Pelican-only" # pragma: no cover

def register(): # for Pelican
    signals.article_generator_finalized.connect(populate_metadata)



#!/usr/bin/env python3
# This file is a part of marzer/poxy and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/poxy/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT
"""
The various entry-point methods used when poxy is invoked from the command line.
"""

import argparse
import datetime
import os
import shutil
import subprocess
import sys
import typing
import zipfile
from pathlib import Path

import colorama
from colorama import Fore, Style

from . import css, doxygen, emoji, graph, mcss, paths
from .run import run
from .schemas import SchemaError
from .utils import *
from .version import *


def _invoker(func, **kwargs):
    colorama.init()
    try:
        func(**kwargs)
    except WarningTreatedAsError as err:
        print(rf'{Style.BRIGHT}{Fore.RED}error:{Style.RESET_ALL} {err} (warning treated as error)', file=sys.stderr)
        sys.exit(1)
    except graph.GraphError as err:
        print_exception(err, include_type=True, include_traceback=True, skip_frames=1)
        sys.exit(1)
    except SchemaError as err:
        print(err, file=sys.stderr)
        sys.exit(1)
    except Error as err:
        print(rf'{Style.BRIGHT}{Fore.RED}error:{Style.RESET_ALL} {err}', file=sys.stderr)
        sys.exit(1)
    except Exception as err:
        print(f'\n{Fore.RED}*************{Style.RESET_ALL}\n', file=sys.stderr)
        print_exception(err, include_type=True, include_traceback=True, skip_frames=1)
        print(
            f'{Fore.RED}*************\n'
            '\nYou appear to have triggered an internal bug!'
            '\nPlease re-run poxy with --bug-report and file an issue at github.com/marzer/poxy/issues'
            '\nMany thanks!'
            f'\n\n*************{Style.RESET_ALL}',
            file=sys.stderr,
        )
        sys.exit(1)
    sys.exit(0)


def make_boolean_optional_arg(args, name, default, help='', **kwargs):
    if sys.version_info >= (3, 9):
        args.add_argument(rf'--{name}', default=default, help=help, action=argparse.BooleanOptionalAction, **kwargs)
    else:
        args.add_argument(rf'--{name}', action=r'store_true', help=help, **kwargs)
        args.add_argument(
            rf'--no-{name}',
            action=r'store_false',
            help=(help if help == argparse.SUPPRESS else None),
            dest=name,
            **kwargs,
        )
        args.set_defaults(**{name: default})


def git(git_args: str, cwd=None) -> typing.Tuple[int, str, str]:
    assert git_args is not None
    proc = subprocess.run(
        ['git'] + str(git_args).strip().split(),
        capture_output=True,
        cwd=str(Path.cwd() if cwd is None else cwd),
        encoding='utf-8',
    )
    return (proc.returncode, proc.stdout.strip() if proc.stdout else "", proc.stderr.strip() if proc.stderr else "")


def git_failed(git_output: typing.Tuple[int, str, str], desc=''):
    assert git_output is not None
    message = rf"git command failed"
    if desc:
        message += rf': {desc}'
    if git_output[0] != 0:  # return code
        message += f'\n  {Style.BRIGHT}exit code:{Style.RESET_ALL}\n    {git_output[0]}'
    if git_output[1]:  # stdout
        message += f'\n  {Style.BRIGHT}stdout:{Style.RESET_ALL}\n'
        message += "\n".join([rf'    {i.strip()}' for i in git_output[1].split('\n') if i.strip()])
    if git_output[2]:  # stderr
        message += f'\n  {Style.BRIGHT}stderr:{Style.RESET_ALL}\n'
        message += "\n".join([rf'    {i.strip()}' for i in git_output[2].split('\n') if i.strip()])
    raise Error(message)


def git_failed_if_nonzero(git_output: typing.Tuple[int, str, str], desc=''):
    if git_output[0] != 0:
        git_failed(git_output, desc)
    return git_output


def multi_version_git_tags(args: argparse.Namespace):
    print('Running in git-tags mode')

    input_dir = args.config.resolve()
    if input_dir.is_file():
        input_dir = input_dir.parent

    if git('diff-files --quiet', cwd=input_dir)[0] or git('diff --no-ext-diff --cached --quiet', cwd=input_dir)[0]:
        raise Error(rf'repository has uncommitted changes')

    print('Fetching...')
    git_failed_if_nonzero(git('fetch --tags', cwd=input_dir))

    current_branch = git_failed_if_nonzero(git('rev-parse --abbrev-ref HEAD', cwd=input_dir))[1]
    print(rf'Current branch: {current_branch}')
    original_branch = current_branch

    default_branch = git_failed_if_nonzero(git('rev-parse --abbrev-ref origin/HEAD', cwd=input_dir))[1]
    default_branch = default_branch.removeprefix(r'origin/')
    print(rf'Default branch: {default_branch}')

    tags = git_failed_if_nonzero(git('tag', cwd=input_dir))[1].splitlines()
    tags = [(t, t.strip().upper().lstrip('V').lstrip()) for t in tags]
    tags = [(t, v) for t, v in tags if v]
    tags = [(t, re.sub(r'\s+', '', v)) for t, v in tags]
    tags = [(t, re.fullmatch(r'[0-9]+([.][0-9]+([.][0-9]+([.][0-9]+)?)?)?', v)) for t, v in tags]
    tags = [(t, v) for t, v in tags if v]
    tags = [(t, v[0].split('.')) for t, v in tags]
    tags = [
        (t, (int(v[0]), int(v[1]) if len(v) > 1 else 0, int(v[2]) if len(v) > 2 else 0, int(v[3]) if len(v) > 3 else 0))
        for t, v in tags
    ]
    tags = sorted(tags, key=lambda t: t[1], reverse=True)
    tags.insert(0, (default_branch, (999999, 999999, 999999, 999999)))

    if args.squash_patches:
        # squash patch/rev differences
        seen_versions = set()
        for i in range(len(tags)):
            normalized_version = (tags[i][1][0], tags[i][1][1])
            if normalized_version in seen_versions:
                tags[i] = None
                continue
            seen_versions.add(normalized_version)
        tags = [t for t in tags if t]

    if args.min_version is not None:
        args.min_version = re.sub(r'[ \t]+', '', str(args.min_version).strip())
        m = re.fullmatch(r'^[vV]?([0-9]+)(?:[.]([0-9]+)(?:[.]([0-9]+)(?:[.]([0-9]+))?)?)?$', args.min_version)
        if m:
            min_ver = (
                (int(m[1] if m[1] else 0)),
                (int(m[2] if m[2] else 0)),
                (int(m[3] if m[3] else 0)),
                (int(m[4] if m[4] else 0)),
            )
            tags = [t for t in tags if t[1] >= min_ver]
        else:
            try:
                max_vers = int(args.min_version)
                assert max_vers < 0
                tags = tags[:-max_vers]
            except:
                raise Error(rf'min-version: expected semver tag or negative integer')

    tags = [t for t, _ in tags]
    print("Versions:")
    print("\n".join([rf'    {t}' for t in tags]))

    worker_args = [
        arg
        for arg in sys.argv[1:]
        if arg
        not in (
            r'--bug-report',
            r'--git-tags',
            r'--worker',
            r'--versions-in-navbar',
            r'--verbose',
            r'-v',
            r'--higest-patch-only',
        )
    ]
    for key in (r'--output-dir', r'--temp-dir', r'--copy-config-to'):
        pos = -1
        try:
            pos = worker_args.index(key)
        except:
            pass
        if pos != -1:
            worker_args.pop(pos)
            worker_args.pop(pos)

    tags_temp_dir = paths.TEMP / 'tags' / temp_dir_name_for(str(Path.cwd()))
    delete_directory(tags_temp_dir)
    tags_temp_dir.mkdir(exist_ok=True, parents=True)

    def cleanup():
        nonlocal current_branch
        nonlocal original_branch
        nonlocal tags_temp_dir
        nonlocal args
        if not args.nocleanup:
            delete_directory(tags_temp_dir)
        if current_branch != original_branch:
            print(rf'Switching back to {Style.BRIGHT}{original_branch}{Style.RESET_ALL}')
            git_failed_if_nonzero(git(rf'checkout {original_branch}', cwd=input_dir))
            current_branch = original_branch

    emitted_tags = set()
    with Defer(cleanup):
        for tag in tags:
            if current_branch != tag:
                print(rf'Checking out {Style.BRIGHT}{tag}{Style.RESET_ALL}')
                git_failed_if_nonzero(git(rf'checkout {tag}', cwd=input_dir))
                current_branch = tag

            if current_branch == default_branch:
                git_failed_if_nonzero(git(rf'pull', cwd=input_dir))

            output_dir = args.output_dir.resolve()
            if tag != default_branch:
                output_dir = tags_temp_dir / tag
                output_dir.mkdir(exist_ok=True, parents=True)

            try:
                print(rf'Generating documentation for {Style.BRIGHT}{tag}{Style.RESET_ALL}')
                result = subprocess.run(
                    args=[
                        r'poxy',
                        r'--worker',
                        r'--versions-in-navbar',
                        r'--output-dir',
                        str(output_dir),
                        *worker_args,
                    ],
                    cwd=str(Path.cwd()),
                    capture_output=not args.verbose,
                    encoding='utf-8',
                )
                if result.returncode != 0:
                    raise Error(
                        rf'Poxy exited with code {result.returncode}{"" if args.verbose else " (re-run with --verbose to see worker output)"}'
                    )
            except Exception as exc:
                msg = rf'documentation generation failed for {Style.BRIGHT}{tag}{Style.RESET_ALL}: {exc}'
                if args.werror:
                    raise WarningTreatedAsError(msg)
                else:
                    print(rf'{Style.BRIGHT}{Fore.YELLOW}warning:{Style.RESET_ALL} {msg}', file=sys.stderr)
                    continue

            if tag != default_branch:
                source_dir = output_dir
                output_dir = args.output_dir.resolve()
                for src in ('html' if args.html else None, 'xml' if args.xml else None):
                    if src is None:
                        continue
                    src = source_dir / src
                    if not src.is_dir():
                        continue
                    dest = output_dir / src.name / tag
                    dest.mkdir(exist_ok=True, parents=True)
                    delete_directory(src / 'poxy')
                    delete_file(src / 'poxy_changelog.html')
                    delete_file(src / 'md_poxy_changelog.html')
                    shutil.copytree(str(src), str(dest), dirs_exist_ok=True)
                    if not args.nocleanup:
                        delete_directory(src)

            emitted_tags.add(tag)

    if not args.html:
        return

    print("Linking versions in HTML output")
    tags = [t for t in tags if t in emitted_tags]
    html_root = args.output_dir.resolve() / 'html'
    for tag in tags:
        html_dir = html_root
        if tag != default_branch:
            html_dir /= tag
        assert_existing_directory(html_dir)
        for fp in get_all_files(html_dir, any=('*.css', '*.html', '*.js'), recursive=False):
            text = read_all_text_from_file(fp)
            if tag != default_branch:
                text = text.replace('href="poxy/', 'href="../poxy/')
                text = text.replace('href="poxy_changelog.html', 'href="../poxy_changelog.html')
                text = text.replace('href="md_poxy_changelog.html', 'href="../md_poxy_changelog.html')
                text = text.replace('src="poxy/', 'src="../poxy/')
            versions = rf'<li class="poxy-navbar-version-selector"><a href="{fp.name}">Version: {"HEAD" if tag == default_branch else tag}</a><ol>'
            for dest_tag in tags:
                target = html_root
                if dest_tag != default_branch:
                    target /= dest_tag
                assert target.is_dir()
                target /= fp.name
                if not target.is_file():
                    target = target.parent / 'index.html'
                assert target.is_file()
                if tag == default_branch and dest_tag != default_branch:
                    target = rf'{dest_tag}/{target.name}'
                elif tag != default_branch and dest_tag == default_branch:
                    target = rf'../{target.name}'
                elif tag != dest_tag:
                    target = rf'../{dest_tag}/{target.name}'
                else:
                    target = target.name
                versions += (
                    rf'<li><a href="{str(target)}">{"HEAD" if dest_tag == default_branch else dest_tag}</a></li>'
                )
            versions += rf'</ol></li>'
            text = re.sub(
                r'<li>\s*<span\s+class="poxy-navbar-version-selector"\s*>\s*FIXME\s*</span>\s*</li>',
                versions,
                text,
                flags=re.I,
            )
            with open(fp, r'w', newline='\n', encoding=r'utf-8') as f:
                f.write(text)


def bug_report(args: argparse.Namespace):
    bug_report_args = [
        arg
        for arg in sys.argv[1:]
        if arg not in (r'--bug-report', r'--worker', r'-v', r'--verbose', r'--keep-original-xml')
    ]
    for key in (r'--output-dir', r'--temp-dir', r'--copy-config-to'):
        pos = -1
        try:
            pos = bug_report_args.index(key)
        except:
            pass
        if pos != -1:
            bug_report_args.pop(pos)
            bug_report_args.pop(pos)

    if '--git-tags' in bug_report_args:
        raise Error(r'--git-tags is currently incompatible with --bug-report. This will be fixed in a later version!')

    bug_report_zip = (Path.cwd() / r'poxy_bug_report.zip').resolve()

    print(r'Preparing output paths')
    delete_directory(paths.BUG_REPORT_DIR)
    delete_file(bug_report_zip)
    paths.BUG_REPORT_DIR.mkdir(exist_ok=True, parents=True)

    print(r'Invoking poxy')
    result = subprocess.run(
        args=[
            r'poxy',
            r'--worker',
            r'--verbose',
            r'--nocleanup',
            r'--keep-original-xml',
            r'--output-dir',
            str(paths.BUG_REPORT_DIR),
            r'--temp-dir',
            str(paths.BUG_REPORT_DIR / r'temp'),
            r'--copy-config-to',
            str(paths.BUG_REPORT_DIR),
            *bug_report_args,
        ],
        cwd=str(Path.cwd()),
        capture_output=True,
        encoding='utf-8',
    )

    if result.stdout:
        print(r'Writing stdout')
        with open(paths.BUG_REPORT_DIR / r'stdout.txt', r'w', newline='\n', encoding=r'utf-8') as f:
            f.write(result.stdout)

    if result.stderr:
        print(r'Writing stderr')
        with open(paths.BUG_REPORT_DIR / r'stderr.txt', r'w', newline='\n', encoding=r'utf-8') as f:
            f.write(result.stderr)

    print(r'Writing metadata')
    with open(paths.BUG_REPORT_DIR / r'metadata.txt', r'w', newline='\n', encoding=r'utf-8') as f:
        f.write(f'version: {VERSION_STRING}\n')
        f.write(f'args: {bug_report_args}\n')
        f.write(f'returncode: {result.returncode}\n')
        try:
            f.write(f'doxygen: {doxygen.raw_version_string()}\n')
        except:
            f.write(f'doxygen: --version failed\n')

    # zip file
    print(r'Zipping files')
    with zipfile.ZipFile(str(bug_report_zip), 'w', compression=zipfile.ZIP_DEFLATED, compresslevel=9) as zip:
        file_prefix_len = len(str(paths.BUG_REPORT_DIR))
        for file in get_all_files(paths.BUG_REPORT_DIR, recursive=True):
            if file.suffix is not None and file.suffix.lower() in (r'.pyc',):
                continue
            relative_file = str(file)[file_prefix_len:].replace('\\', '/').strip('/')
            zip.write(file, arcname=rf'poxy_bug_report/{relative_file}')

    print(r'Cleaning up')
    delete_directory(paths.BUG_REPORT_DIR)

    print(
        f'Zip generated: {Style.BRIGHT}{bug_report_zip}{Style.RESET_ALL}\n'
        rf'{Fore.CYAN}{Style.BRIGHT}Please attach this file when you make a report at https://github.com/marzer/poxy/issues, thanks!{Style.RESET_ALL}'
    )


def main(invoker=True):
    """
    The entry point when the library is invoked as `poxy`.
    """
    if invoker:
        _invoker(main, invoker=False)
        return

    # yapf: disable
    args = argparse.ArgumentParser(
        description=
        rf'{Fore.CYAN}{Style.BRIGHT}'
        r'''  _ __   _____  ___   _ ''' '\n'
        r''' | '_ \ / _ \ \/ / | | |''' '\n'
        r''' | |_) | (_) >  <| |_| |''' '\n'
        r''' | .__/ \___/_/\_\\__, |''' '\n'
        r''' | |               __/ |''' '\n'
        r''' |_|              |___/ ''' rf'{Style.RESET_ALL} v{VERSION_STRING} - github.com/marzer/poxy'
        '\n\n'
        r'Generate fancy C++ documentation.',
        formatter_class=argparse.RawTextHelpFormatter
    )
    # yapf: enable

    # --------------------------------------------------------------
    # public user-facing arguments
    # --------------------------------------------------------------

    args.add_argument(
        r'config',
        type=Path,
        nargs='?',
        default=Path.cwd(),
        help=r'path to poxy.toml or a directory containing it (default: .)',
    )
    args.add_argument(r'-v', r'--verbose', action=r'store_true', help=r"enable very noisy diagnostic output")  #
    make_boolean_optional_arg(args, r'html', default=True, help=r'specify whether HTML output is required')  #
    args.add_argument(
        r'--ppinclude',  #
        type=str,
        default=None,
        metavar=r'<regex>',
        help=r"pattern matching HTML file names to post-process (default: all)",
    )
    args.add_argument(
        r'--ppexclude',  #
        type=str,
        default=None,
        metavar=r'<regex>',
        help=r"pattern matching HTML file names to exclude from post-processing (default: %(default)s)",
    )
    args.add_argument(
        r'--theme',  #
        choices=[r'light', r'dark', r'custom'],
        default=None,
        help=r'sets the default visual theme (default: read from config)',
    )
    args.add_argument(
        r'--threads',
        type=int,
        default=0,
        metavar=r'N',
        help=r"set the number of threads to use (default: automatic)",  #
    )
    args.add_argument(r'--version', action=r'store_true', help=r"print the version and exit", dest=r'print_version')  #
    make_boolean_optional_arg(args, r'xml', default=False, help=r'specify whether XML output is required')  #
    make_boolean_optional_arg(
        args, r'werror', default=None, help=r'treat warnings as errors (default: read from config)'
    )  #
    args.add_argument(
        r'--bug-report', action=r'store_true', help=r"captures all output in a zip file for easier bug reporting."  #
    )
    args.add_argument(
        r'--git-tags', action=r'store_true', help=r"add git-tag-based semver version switcher to the generated HTML"  #
    )
    make_boolean_optional_arg(
        args,
        r'squash-patches',
        default=True,
        help='when using --git-tags and two version tags differ by a patch number,\ngenerate docs for the highest one only (default: %(default)s)',
    )
    args.add_argument(
        r'--min-version',
        type=str,
        default=None,
        metavar='<version>',
        help='sets the minimum version number to emit when using --git-tags,\nor a negative integer to mean "the last N versions". (default: %(default)s)',
    )

    # --------------------------------------------------------------
    # hidden/developer-only/deprecated/diagnostic arguments
    # --------------------------------------------------------------

    args.add_argument(r'--where', action=r'store_true', help=argparse.SUPPRESS)  #
    args.add_argument(r'--nocleanup', action=r'store_true', help=argparse.SUPPRESS)  #
    args.add_argument(r'--noassets', action=r'store_true', help=argparse.SUPPRESS)  #
    args.add_argument(r'--update-styles', action=r'store_true', help=argparse.SUPPRESS)  #
    make_boolean_optional_arg(args, r'update-fonts', default=None, help=argparse.SUPPRESS)
    args.add_argument(r'--update-emoji', action=r'store_true', help=argparse.SUPPRESS)  #
    args.add_argument(r'--update-tests', action=r'store_true', help=argparse.SUPPRESS)  #
    args.add_argument(r'--doxygen-version', action=r'store_true', help=argparse.SUPPRESS)  #
    args.add_argument(r'--update-mcss', type=Path, default=None, help=argparse.SUPPRESS)  #
    args.add_argument(  # --xml and --html are the replacements for --xmlonly
        r'--xmlonly', action=r'store_true', help=argparse.SUPPRESS  #
    )
    args.add_argument(r'--xml-v2', action=r'store_true', help=argparse.SUPPRESS)  #
    args.add_argument(r'--worker', action=r'store_true', help=argparse.SUPPRESS)  #
    args.add_argument(r'--output-dir', type=Path, default=Path.cwd(), help=argparse.SUPPRESS)  #
    args.add_argument(r'--temp-dir', type=Path, default=None, help=argparse.SUPPRESS)  #
    args.add_argument(r'--copy-config-to', type=Path, default=None, help=argparse.SUPPRESS)  #
    args.add_argument(r'--versions-in-navbar', action=r'store_true', help=argparse.SUPPRESS)  #
    args.add_argument(r'--keep-original-xml', action=r'store_true', help=argparse.SUPPRESS)  #
    args = args.parse_args()

    # --------------------------------------------------------------
    # --version
    # --------------------------------------------------------------

    if args.print_version:
        print(VERSION_STRING)
        return

    # --------------------------------------------------------------
    # --doxygen-version
    # --------------------------------------------------------------

    if args.doxygen_version:
        print(doxygen.version_string())
        return

    # --------------------------------------------------------------
    # --where
    # --------------------------------------------------------------

    if args.where:
        print(paths.PACKAGE)
        return

    # --------------------------------------------------------------
    # developer-only subcommands
    # --------------------------------------------------------------

    if args.update_mcss is not None:
        args.update_styles = True
        if args.update_fonts is None:
            args.update_fonts = True
        mcss.update_bundled_install(args.update_mcss)
    assert_existing_directory(paths.MCSS)
    assert_existing_file(Path(paths.MCSS, r'documentation/doxygen.py'))

    if args.update_fonts:
        args.update_styles = True
    if args.update_styles:
        css.regenerate_builtin_styles(use_cached_fonts=not args.update_fonts)

    if args.update_emoji:
        emoji.update_database_file()

    if args.update_tests:
        if not paths.TESTS.exists() or not paths.TESTS.is_dir():
            raise Exception(
                f'{paths.TESTS} did not exist or was not a directory (--update-tests is only for editable installations)'
            )
        run_python_script(
            Path(paths.TESTS, r'regenerate_tests.py'),  #
            *[a for a in (r'--verbose' if args.verbose else None, r'--nocleanup' if args.nocleanup else None) if a],
        )

    if (
        args.update_styles
        or args.update_fonts
        or args.update_emoji
        or args.update_tests
        or args.update_mcss is not None
    ):
        return

    if not args.worker:
        print(rf'{Fore.CYAN}{Style.BRIGHT}poxy{Style.RESET_ALL} v{VERSION_STRING}')

    # --------------------------------------------------------------
    # bug report invocation
    # --------------------------------------------------------------

    if args.bug_report:
        bug_report(args)
        return

    # --------------------------------------------------------------
    # multi-version invocation
    # --------------------------------------------------------------

    if args.git_tags:
        with ScopeTimer(r'All tasks', print_start=False, print_end=True) as timer:
            multi_version_git_tags(args)
        return

    # --------------------------------------------------------------
    # regular invocation
    # --------------------------------------------------------------

    if args.xmlonly:
        args.html = False
        args.xml = True

    with ScopeTimer(r'All tasks', print_start=False, print_end=not args.worker) as timer:
        run(
            # named args:
            config_path=args.config,
            output_dir=args.output_dir,
            output_html=args.html,
            output_xml=args.xml,
            threads=args.threads,
            cleanup=not args.nocleanup,
            verbose=args.verbose,
            logger=True,  # stderr + stdout
            html_include=args.ppinclude,
            html_exclude=args.ppexclude,
            treat_warnings_as_errors=args.werror,
            theme=args.theme,
            copy_assets=not args.noassets,
            temp_dir=args.temp_dir,
            copy_config_to=args.copy_config_to,
            versions_in_navbar=args.versions_in_navbar,
            keep_original_xml=args.keep_original_xml,
            # kwargs:
            xml_v2=args.xml_v2,
        )


def main_blog_post(invoker=True):
    """
    The entry point when the library is invoked as `poxyblog`.
    """
    if invoker:
        _invoker(main_blog_post, invoker=False)
        return

    args = argparse.ArgumentParser(
        description=r'Initializes a new blog post for Poxy sites.', formatter_class=argparse.RawTextHelpFormatter
    )
    args.add_argument(r'title', type=str, help=r'the title of the new blog post')  #
    args.add_argument(r'-v', r'--verbose', action=r'store_true', help=r"enable very noisy diagnostic output")  #
    args.add_argument(r'--version', action=r'store_true', help=r"print the version and exit", dest=r'print_version')  #
    args = args.parse_args()

    if args.print_version:
        print(VERSION_STRING)
        return

    print(rf'{Fore.CYAN}{Style.BRIGHT}poxy{Style.RESET_ALL} v{VERSION_STRING}')

    date = datetime.datetime.now().date()

    title = args.title.strip()
    if not title:
        raise Error(r'title cannot be blank.')
    if re.search(r''''[\n\v\f\r]''', title) is not None:
        raise Error(r'title cannot contain newline characters.')
    file = re.sub(r'''[!@#$%^&*;:'"<>?/\\\s|+]+''', '_', title)
    file = rf'{date}_{file.lower()}.md'

    blog_dir = Path(r'blog')
    if blog_dir.exists() and not blog_dir.is_dir():
        raise Error(rf'{blog_dir.resolve()} already exists and is not a directory')
    blog_dir.mkdir(exist_ok=True)

    file = Path(blog_dir, file)
    if file.exists():
        if not file.is_file():
            raise Error(rf'{file.resolve()} already exist and is not a file')
        raise Error(rf'{file.resolve()} already exists')

    with open(file, r'w', encoding=r'utf-8', newline='\n') as f:
        write = lambda s='', end='\n': print(s, end=end, file=f)
        write(rf'# {title}')
        write()
        write()
        write()
        write()
        write()
        write(r'<!--[poxy_metadata[')
        write(rf'tags = []')
        write(r']]-->')
    print(rf'Blog post file initialized: {file.resolve()}')


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# This file is a part of marzer/poxy and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/poxy/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT
"""
Functions and classes for working with m.css.
"""

import shutil

from . import paths
from .utils import *


def update_bundled_install(source_root: Path):
    assert source_root is not None

    source_root = coerce_path(source_root).resolve()
    assert_existing_directory(source_root)
    assert_existing_file(Path(source_root, r'documentation/doxygen.py'))
    if paths.MCSS == source_root:
        raise Exception(r'm.css source path may not be the same as the internal destination.')

    # delete existing m.css
    if paths.MCSS.exists():
        delete_directory(paths.MCSS, logger=True)

    # copy new one
    print(rf'Updating bundled m.css from {source_root}')
    shutil.copytree(
        source_root,
        paths.MCSS,
        symlinks=False,
        dirs_exist_ok=True,
        ignore=shutil.ignore_patterns(
            r'.git*',  #
            r'.editor*',
            r'.circleci*',
            r'.coverage*',
            r'.istanbul*',
            r'*.idx',
            r'*.pyc',
            r'*.compiled.css',
            r'__pycache__*',
            r'artwork*',
            r'circleci*',
            r'test_doxygen*',
            r'test_python*',
            r'pelican-theme*',
            r'pygments-*.py',
            r'postprocess.sh',
            r'postprocess.py',
            r'm-*dark.css',  # the m.css themes have local copies in data/ for *reasons*
            r'm-*light.css',
        ),
    )

    # delete unwanted files
    # todo: figure how to do this as a filter, instead of copying -> deleting
    for folder in (
        r'artwork',
        r'doc',
        r'documentation/test',
        r'documentation/templates/python',
        r'package',
        r'pelican-theme',
        r'plugins/m/test',
        r'site',
    ):
        delete_directory(Path(paths.MCSS, folder), logger=True)


__all__ = ['update_bundled_install']



#
#   This file is part of m.css.
#
#   Copyright © 2017, 2018, 2019, 2020, 2021, 2022, 2023, 2024
#             Vladimír Vondruš <mosra@centrum.cz>
#
#   Permission is hereby granted, free of charge, to any person obtaining a
#   copy of this software and associated documentation files (the "Software"),
#   to deal in the Software without restriction, including without limitation
#   the rights to use, copy, modify, merge, publish, distribute, sublicense,
#   and/or sell copies of the Software, and to permit persons to whom the
#   Software is furnished to do so, subject to the following conditions:
#
#   The above copyright notice and this permission notice shall be included
#   in all copies or substantial portions of the Software.
#
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#   DEALINGS IN THE SOFTWARE.
#

import re
import subprocess

from docutils import nodes
from docutils.parsers import rst
from docutils.parsers.rst import directives
from docutils.parsers.rst.roles import set_classes

try:
    import dot2svg
except ImportError:
    # While the above was enough to make things work with Pelican 4.2 and
    # before (and also works with other m.css tools like the Python doc
    # generator), Pelican 4.5.1+ needs the below (4.5.0 didn't work with
    # namespaced plugins at all). See the comment in m.code for further info.
    import sys
    sys.path.append(os.path.join(os.path.dirname(__file__), '..'))
    import dot2svg

def _is_graph_figure(parent):
    # The parent has to be a figure, marked as m-figure
    if not isinstance(parent, nodes.figure): return False
    if 'm-figure' not in parent.get('classes', []): return False

    # And as a first visible node of such type
    for child in parent:
        if not isinstance(child, nodes.Invisible): return False

    return True

class Dot(rst.Directive):
    has_content = True
    optional_arguments = 1
    final_argument_whitespace = True
    option_spec = {'class': directives.class_option,
                   'name': directives.unchanged}

    def run(self, source):
        set_classes(self.options)

        # If this is the first real node inside a graph figure, put the SVG
        # directly inside
        parent = self.state.parent
        if _is_graph_figure(parent):
            svg = dot2svg.dot2svg(source, attribs=' class="{}"'.format(' '.join(['m-graph'] + self.options.get('classes', []))))
            node = nodes.raw('', svg, format='html')
            return [node]

        # Otherwise wrap it in a <div class="m-graph">
        svg = dot2svg.dot2svg(source)
        container = nodes.container(**self.options)
        container['classes'] = ['m-graph'] + container['classes']
        node = nodes.raw('', svg, format='html')
        container.append(node)
        return [container]

class Digraph(Dot):
    def run(self):
        # We need to pass "" for an empty title to get rid of <title>,
        # otherwise the output contains <title>%3</title> (wtf!)
        return Dot.run(self, 'digraph "{}" {{\n{}}}'.format(
            self.arguments[0] if self.arguments else '',
            '\n'.join(self.content)))

class StrictDigraph(Dot):
    def run(self):
        # We need to pass "" for an empty title to get rid of <title>,
        # otherwise the output contains <title>%3</title> (wtf!)
        return Dot.run(self, 'strict digraph "{}" {{\n{}}}'.format(
            self.arguments[0] if self.arguments else '',
            '\n'.join(self.content)))

class Graph(Dot):
    def run(self):
        # We need to pass "" for an empty title to get rid of <title>,
        # otherwise the output contains <title>%3</title> (wtf!)
        return Dot.run(self, 'graph "{}" {{\n{}}}'.format(
            self.arguments[0] if self.arguments else '',
            '\n'.join(self.content)))

class StrictGraph(Dot):
    def run(self):
        # We need to pass "" for an empty title to get rid of <title>,
        # otherwise the output contains <title>%3</title> (wtf!)
        return Dot.run(self, 'strict graph "{}" {{\n{}}}'.format(
            self.arguments[0] if self.arguments else '',
            '\n'.join(self.content)))

def register_mcss(mcss_settings, **kwargs):
    dot2svg.configure(
        mcss_settings.get('M_DOT_FONT', 'Source Sans Pro'),
        mcss_settings.get('M_DOT_FONT_SIZE', 16.0))
    rst.directives.register_directive('digraph', Digraph)
    rst.directives.register_directive('strict-digraph', StrictDigraph)
    rst.directives.register_directive('graph', Graph)
    rst.directives.register_directive('strict-graph', StrictGraph)

# Below is only Pelican-specific functionality. If Pelican is not found, these
# do nothing.

def _pelican_configure(pelicanobj):
    register_mcss(mcss_settings=pelicanobj.settings)

def register(): # for Pelican
    from pelican import signals

    signals.initialized.connect(_pelican_configure)



#!/usr/bin/env python3
# This file is a part of marzer/poxy and is subject to the the terms of the MIT license.
# Copyright (c) Mark Gillard <mark.gillard@outlook.com.au>
# See https://github.com/marzer/poxy/blob/master/LICENSE for the full license text.
# SPDX-License-Identifier: MIT
"""
'Fixer' function objects used during various post-process steps.
"""

import html

from bs4 import NavigableString
from trieregex import TrieRegEx
from typing import Tuple, Union

from . import soup
from .project import Context
from .svg import SVG
from .utils import *

# =======================================================================================================================
# base classes
# =======================================================================================================================


class HTMLFixer(object):
    pass


class PlainTextFixer(object):
    pass


# =======================================================================================================================
# HTML post-processes
# =======================================================================================================================

# yapf: disable

PAIRED_TAGS = TrieRegEx(
    r'aside',
    r'code',
    r'div',
    r'p',
    r'pre',
    r'span',
    r'aside',
    r'b',
    r'center',
    r'em',
    r'h1',
    r'h2',
    r'h3',
    r'h4',
    r'h5',
    r'h6',
    r'i',
    r'li',
    r'ol',
    r'strong',
    r'u',
    r'ul',
)
PAIRED_TAGS = PAIRED_TAGS.regex()
PAIRED_TAGS = re.compile(
    r'\[\s*'
    + rf'({PAIRED_TAGS})\s*'  # group 1: tag name
    + r'([^\]]*?)\s*'  # group 2: tag attributes
    + r'\]'
    + r'(.*?)'  # group 3: tag content
    + r'\[\s*/\s*\1\s*\]',  # closer
    re.I | re.S,
)

SINGLE_TAGS = TrieRegEx(
    r'add_class',
    r'add_parent_class',
    r'add_parent_parent_class',
    r'htmlentity',
    r'entity',
    r'emoji',
    r'img',
    r'li',
    r'ol',
    r'parent_add_class',
    r'parent_parent_add_class',
    r'parent_parent_remove_class',
    r'parent_parent_set_class',
    r'parent_parent_set_id',
    r'parent_parent_set_name',
    r'parent_remove_class',
    r'parent_set_class',
    r'parent_set_id',
    r'parent_set_name',
    r'remove_class',
    r'remove_parent_class',
    r'remove_parent_parent_class',
    r'set_class',
    r'set_id',
    r'set_name',
    r'set_parent_class',
    r'set_parent_class',
    r'set_parent_id',
    r'set_parent_id',
    r'set_parent_name',
    r'set_parent_name',
    r'ul',
)
SINGLE_TAGS = SINGLE_TAGS.regex()
SINGLE_TAGS = re.compile(
    r'\[\s*'
    + rf'({SINGLE_TAGS})\s*'  # group 1: tag name
    + r'([^\]]*?)\s*'  # group 2: tag attributes
    + r'\]',
    re.I | re.S,
)

# yapf: enable

TAG_PARENTS = TrieRegEx(
    r'dd',
    r'p',
    r'h1',
    r'h2',
    r'h3',
    r'h4',
    r'h5',
    r'h6',
    r'li',
    r'aside',
    r'td',
    r'div',
    r'span',
    r'a',
    r'i',
    r'u',
    r'b',
)
TAG_PARENTS = TAG_PARENTS.regex()
TAG_PARENTS = re.compile(rf'^{TAG_PARENTS}$', re.I)

TAG_DISALLOWED_PARENTS = (r'code', r'pre')


class CustomTags(HTMLFixer):
    '''
    Modifies HTML using custom square-bracket [tags].
    '''

    __hex_entity = re.compile(r'(?:[0#]?[xX])?([a-fA-F0-9]+)')

    def __paired_tags_substitute(cls, m, out, context):
        tag_name = m[1].lower()
        tag_attrs = m[2].strip() if m[2] else ''
        tag_attrs = rf' {tag_attrs}' if tag_attrs else ''
        tag_content = m[3].strip() if m[3] else ''
        if tag_content:
            return rf'<{tag_name}{tag_attrs}>{tag_content}</{tag_name}>'
        return rf'<{tag_name}{tag_attrs} />'

    @classmethod
    def __single_tags_substitute(cls, m, out, context):
        tag_name = m[1].lower()
        tag_attrs = m[2].strip() if m[2] else ''
        if tag_name in (r'htmlentity', r'entity'):
            if not tag_attrs:
                return ''
            hex_match = cls.__hex_entity.fullmatch(tag_attrs)
            if hex_match:
                try:
                    cp = int(hex_match[1], 16)
                    if cp <= 0x10FFFF:
                        return rf'&#x{hex_match[1]};'
                except:
                    pass
            return f'&{tag_attrs};'
        elif tag_name == r'emoji':
            if not tag_attrs:
                return ''
            tag_attrs = tag_attrs.lower()
            emoji = None
            for base in (16, 10):
                try:
                    emoji = context.emoji[int(tag_attrs, base)]
                    if emoji is not None:
                        break
                except:
                    pass
            if emoji is None:
                emoji = context.emoji[tag_attrs]
            return str(emoji) if emoji is not None else ''
        elif tag_name.find(r'_class') != -1:
            if tag_attrs:
                tag_attrs = [s.strip() for s in tag_attrs.split()]
                tag_attrs = [s for s in tag_attrs if s]
                if tag_attrs:
                    out.append((tag_name, tag_attrs))
            return ''
        elif tag_name.find(r'_name') != -1 or tag_name.find(r'_id') != -1:
            if tag_attrs:
                out.append((tag_name, tag_attrs))
            return ''
        else:
            return rf'<{tag_name}{rf" {tag_attrs}" if tag_attrs else ""}>'

    def __call__(self, context: Context, doc: soup.HTMLDocument, path: Path):
        if doc.article_content is None:
            return False

        changed = False

        def get_candidate_tags():
            tags = doc.article_content.find_all(TAG_PARENTS)
            tags = [tag for tag in tags if not tag.decomposed and len(tag.contents)]
            tags = [tag for tag in tags if soup.find_parent(tag, TAG_DISALLOWED_PARENTS, doc.article_content) is None]
            return tags

        # paired tags
        changed_this_pass = True
        while changed_this_pass:
            changed_this_pass = False
            for tag in get_candidate_tags():
                replacer = RegexReplacer(
                    PAIRED_TAGS, lambda m, out: self.__paired_tags_substitute(m, out, context), str(tag)
                )
                if replacer:
                    changed_this_pass = True
                    soup.replace_tag(tag, str(replacer))
                    break
            if changed_this_pass:
                doc.smooth()
                changed = True

        # single tags
        changed_this_pass = True
        while changed_this_pass:
            changed_this_pass = False
            tags = get_candidate_tags()
            strings = []
            for tag in tags:
                strings += [string for string in tag.children if isinstance(string, NavigableString) and len(string)]
            for string in strings:
                before = str(string)
                replacer = RegexReplacer(
                    SINGLE_TAGS, lambda m, out: self.__single_tags_substitute(m, out, context), str(string)
                )
                if replacer:
                    changed_this_pass = True
                    parent = string.parent
                    new_tags = soup.replace_tag(string, str(replacer))
                    if parent is not None and parent.name == 'p' and not len(parent.contents):
                        parent = parent.parent
                    for i in range(len(replacer)):  # custom tag handling
                        key = replacer[i][0]
                        if key.find(r'parent_') != -1:
                            if key.find(r'parent_parent') != -1:
                                key = key.replace(r'parent_parent', r'parent')
                                if parent is not None:
                                    parent = parent.parent
                            if parent is None:
                                continue
                            if key in (r'parent_add_class', r'add_parent_class'):
                                soup.add_class(parent, replacer[i][1])
                            elif key in (r'parent_remove_class', r'remove_parent_class'):
                                soup.remove_class(parent, replacer[i][1])
                            elif key in (r'parent_set_class', r'set_parent_class'):
                                soup.set_class(parent, replacer[i][1])
                            elif key in (r'parent_set_name', r'set_parent_name'):
                                parent.name = replacer[i][1]
                            elif key in (r'parent_set_id', r'set_parent_id'):
                                parent['id'] = replacer[i][1]
                        elif key.find(r'_class') or key.find(r'_name') != -1 or key.find(r'_id') != -1:
                            target = None
                            if len(new_tags) == 1:
                                target = new_tags[0]
                            elif not new_tags:
                                target = parent
                            if target is not None and isinstance(target, NavigableString):
                                target = target.parent
                            if not target:
                                continue
                            if key == r'add_class':
                                soup.add_class(target, replacer[i][1])
                            elif key == r'remove_class':
                                soup.remove_class(target, replacer[i][1])
                            elif key == r'set_class':
                                soup.set_class(target, replacer[i][1])
                            elif key == r'set_name':
                                target.name = replacer[i][1]
                            elif key == r'set_id':
                                target.id = replacer[i][1]
                    continue
            if changed_this_pass:
                doc.smooth()
                changed = True

        return changed


class _CPPModifiersBase(HTMLFixer):
    '''
    Base type for modifier parsing fixers.
    '''

    _modifierRegex = r"defaulted|noexcept|constexpr|(?:pure )?virtual|protected|__(?:(?:vector|std|fast)call|cdecl)"
    _modifierClasses = {
        "defaulted": "m-info",
        "noexcept": "m-success",
        "constexpr": "m-primary",
        "pure virtual": "m-warning",
        "virtual": "m-warning",
        "protected": "m-warning",
        "__vectorcall": "m-special",
        "__stdcall": "m-special",
        "__fastcall": "m-special",
        "__cdecl": "m-special",
    }


class CPPModifiers1(_CPPModifiersBase):
    '''
    Fixes improperly-parsed modifiers on function signatures in the various 'detail view' sections.
    '''

    __expression = re.compile(rf'(\s+)({_CPPModifiersBase._modifierRegex})(\s+)')
    __sections = ('pub-static-methods', 'pub-methods', 'friends', 'func-members')

    @classmethod
    def __substitute(cls, m, out):
        return f'{m[1]}<span class="poxy-injected m-label m-flat {cls._modifierClasses[m[2]]}">{m[2]}</span>{m[3]}'

    def __call__(self, context: Context, doc: soup.HTMLDocument, path: Path):
        if doc.article_content is None:
            return False
        changed = False
        for sect in self.__sections:
            tags = doc.find_all_from_sections('dt', select='span.m-doc-wrap', section=sect)
            for tag in tags:
                replacer = RegexReplacer(self.__expression, self.__substitute, str(tag))
                if replacer:
                    changed = True
                    soup.replace_tag(tag, str(replacer))
        return changed


class CPPModifiers2(_CPPModifiersBase):
    '''
    Fixes improperly-parsed modifiers on function signatures in the 'Function documentation' section.
    '''

    __expression = re.compile(rf'\s+({_CPPModifiersBase._modifierRegex})\s+')

    @classmethod
    def __substitute(cls, m, matches):
        matches.append(m[1])
        return ' '

    def __call__(self, context: Context, doc: soup.HTMLDocument, path: Path):
        if doc.article_content is None:
            return False
        changed = False
        sections = doc.find_all_from_sections(section=False)  # all sections without an id
        section = None
        for s in sections:
            if str(s.h2.string) == 'Function documentation':
                section = s
                break
        if section is not None:
            funcs = section(id=True)
            funcs = [f.find('h3') for f in funcs]
            for f in funcs:
                bumper = f.select_one('span.m-doc-wrap-bumper')
                end = f.select_one('span.m-doc-wrap').contents
                end = end[len(end) - 1]
                matches = []
                bumperContent = self.__expression.sub(lambda m: self.__substitute(m, matches), str(bumper))
                if matches:
                    changed = True
                    soup.replace_tag(bumper, bumperContent)
                    lastInserted = end.find('span')
                    for match in matches:
                        lastInserted = doc.new_tag(
                            'span',
                            parent=end,
                            string=match,
                            class_=f'poxy-injected m-label {self._modifierClasses[match]}',
                            before=lastInserted,
                        )
                        lastInserted.insert_after(' ')
        return changed


class StripIncludes(HTMLFixer):
    '''
    Strips #include <paths/to/headers.h> based on context.sources.strip_includes.
    '''

    def __call__(self, context: Context, doc: soup.HTMLDocument, path: Path):
        if doc.article is None or not context.sources.strip_includes:
            return False
        changed = False
        for include_div in doc.article.find_all(r'div', class_=r'm-doc-include'):
            anchor = include_div.find('a', href=True, class_=r'cpf')
            if anchor is None:
                continue
            text = anchor.get_text()
            if not (text.startswith('<') and text.endswith('>')):
                continue
            text = text[1:-1].strip()
            for strip in context.sources.strip_includes:
                if len(text) < len(strip) or not text.startswith(strip):
                    continue
                if len(text) == len(strip):
                    soup.destroy_node(include_div)
                else:
                    anchor.contents.clear()
                    anchor.contents.append(NavigableString(rf'<{text[len(strip):]}>'))
                changed = True
                break
        return changed


class Banner(HTMLFixer):
    '''
    Makes the first image on index.html a 'banner'
    '''

    def __call__(self, context: Context, doc: soup.HTMLDocument, path: Path):
        if doc.article_content is None or path.name.lower() != 'index.html':
            return False
        parent = doc.article_content

        h1 = parent.find('h1', recursive=False)
        banner = parent.find('img', recursive=False)
        if not banner or not h1 or r'src' not in banner.attrs or not banner[r'src']:
            return False

        # ensure it's the first image in the page, before any subsections or headings
        for sibling_tag in ('section', 'h2', 'h3', 'h4', 'h5', 'h6'):
            sibling = banner.find_previous_sibling(sibling_tag)
            if sibling is not None:
                return False

        banner = banner.extract()
        h1.replace_with(banner)
        banner[r'id'] = r'poxy-main-banner'
        soup.add_class(doc.body, r'poxy-has-main-banner')

        if context.badges:
            soup.add_class(doc.body, r'poxy-has-badges')
            span_size = 0
            span_size = 2 if not span_size and (len(context.badges) % 2) == 0 else span_size
            span_size = 3 if not span_size and (len(context.badges) % 3) == 0 else span_size
            span_size = 2 if not span_size else span_size
            idx = 0
            span = None
            parent = doc.new_tag(r'div', id=r'poxy-badges', after=banner)
            for alt, src, href in context.badges:
                if span is None or (idx % span_size) == 0:
                    span = doc.new_tag(r'span', parent=parent)
                img_parent = span
                if href:
                    img_parent = doc.new_tag(r'a', parent=span, href=href, target=r'_blank')
                img = doc.new_tag(r'img', parent=img_parent, src=src)
                if alt:
                    img[r'alt'] = alt
                idx += 1
        return True


class CodeBlocks(HTMLFixer):
    '''
    Fixes various issues and improves syntax highlighting in <code> blocks.
    '''

    __keywords = (
        r'alignas',
        r'alignof',
        r'bool',
        r'char',
        r'char16_t',
        r'char32_t',
        r'char8_t',
        r'class',
        r'const',
        r'consteval',
        r'constexpr',
        r'constinit',
        r'do',
        r'double',
        r'else',
        r'explicit',
        r'false',
        r'float',
        r'if',
        r'inline',
        r'int',
        r'long',
        r'mutable',
        r'noexcept',
        r'short',
        r'signed',
        r'sizeof',
        r'struct',
        r'template',
        r'true',
        r'typename',
        r'unsigned',
        r'void',
        r'wchar_t',
        r'while',
    )

    __ns_token_expr = re.compile(r'(?:::|[a-zA-Z_][a-zA-Z_0-9]*|::[a-zA-Z_][a-zA-Z_0-9]*|[a-zA-Z_][a-zA-Z_0-9]*::)')
    __ns_full_expr = re.compile(r'(?:::)?[a-zA-Z_][a-zA-Z_0-9]*(::[a-zA-Z_][a-zA-Z_0-9]*)*(?:::)?')
    __compound_starter_classes = (  # must not contain: fm, o, p, nc, mi, nf, nn
        r'n',
        r'no',
        r'nl',
        r'ne',
        r'nx',
        r'kt',
        r'kr',
        r'nb',
    )
    __compound_classes = (*__compound_starter_classes, r'mi', r'nf', r'nc', r'nn')  # must not contain:  fm, o, p
    __func_name = re.compile(r'^\s*[a-zA-Z_][a-zA-Z0-9_]*\s*$')
    __func_bracket = re.compile(r'^\s*[(]')

    @classmethod
    def __colourize_compound_def(cls, tags, context) -> bool:
        assert tags
        assert tags[0].string != '::'
        assert len(tags) == 1 or tags[-1].string != '::'
        full_str = ''.join([tag.get_text() for tag in tags])

        def colourize_case(c: str) -> bool:
            nonlocal cls
            nonlocal tags
            changed = False
            if soup.get_classes(tags[-1]) != [c]:
                soup.set_class(tags[-1], c)
                changed = True
            del tags[-1]
            while tags and tags[-1].string == r'::':
                del tags[-1]
            if tags:
                changed = cls.__colourize_compound_def(tags, context) or changed
            return changed

        if context.code_blocks.enums.fullmatch(full_str):
            return colourize_case(r'mi')  # Literal.Number.Integer

        if context.code_blocks.functions.fullmatch(full_str):
            return colourize_case(r'nf')  # Name.Function

        if context.code_blocks.types.fullmatch(full_str):
            return colourize_case(r'nc')  # Name.Class

        while not context.code_blocks.namespaces.fullmatch(full_str):
            del tags[-1]
            while tags and tags[-1].string == r'::':
                del tags[-1]
            if not tags:
                break
            full_str = ''.join([tag.get_text() for tag in tags])

        if tags:
            changed = False
            while len(tags) > 1:
                tags.pop(-1).decompose()
                changed = True
            changed = changed or tags[-1].string != full_str
            tags[-1].string = full_str
            if soup.get_classes(tags[-1]) != [r'nn']:  # Name.Namespace
                soup.set_class(tags[-1], r'nn')
                return True
            return changed

        return False

    def __call__(self, context: Context, doc: soup.HTMLDocument, path: Path):
        changed = False

        # fix up syntax highlighting
        code_blocks = doc.body(('pre', 'code'), class_='m-code')
        changed_this_pass = True
        while changed_this_pass:
            changed_this_pass = False
            for code_block in code_blocks:
                changed_this_block = False

                # c-style multi-line comments (doxygen butchers them)
                mlc_open = code_block.find('span', class_='o', string='/!*')
                while mlc_open is not None:
                    mlc_close = mlc_open.find_next_sibling('span', class_='o', string='*!/')
                    if mlc_close is None:
                        break
                    changed_this_block = True
                    next_open = mlc_close.find_next_sibling('span', class_='o', string='/!*')

                    tags = []
                    current = mlc_open
                    while current is not None:
                        tags.append(current)
                        if current is mlc_close:
                            break
                        current = current.next_sibling

                    mlc_open.string = '/*'
                    mlc_close.string = '*/'
                    string = ''
                    for tag in tags:
                        string = string + tag.get_text()
                    mlc_open.string = string
                    soup.set_class(mlc_open, 'cm')
                    while len(tags) > 1:
                        soup.destroy_node(tags.pop())

                    mlc_open = next_open

                # macros
                spans = code_block(r'span', class_=self.__compound_classes, string=True)
                for span in spans:
                    if context.code_blocks.macros.fullmatch(span.get_text()):
                        soup.set_class(span, r'fm')  # Name.Function.Magic
                        changed_this_block = True

                if 1:
                    # collect all names and glom them all together as compound names
                    spans = code_block(r'span', class_=self.__compound_starter_classes, string=True)
                    compound_names = []
                    compound_name_evaluated_tags = set()
                    for i in range(0, len(spans)):
                        current = spans[i]
                        if id(current) in compound_name_evaluated_tags:
                            continue

                        compound_name_evaluated_tags.add(id(current))
                        tags = [current]
                        while True:
                            prev = current.previous_sibling
                            if (
                                prev is None
                                or prev.string is None
                                or isinstance(prev, NavigableString)
                                or not soup.has_any_classes(prev, *self.__compound_classes, r'o', r'p')
                                or not self.__ns_token_expr.fullmatch(prev.string)
                            ):
                                break
                            current = prev
                            tags.insert(0, current)
                            compound_name_evaluated_tags.add(id(current))

                        current = spans[i]
                        while True:
                            nxt = current.next_sibling
                            if (
                                nxt is None
                                or nxt.string is None
                                or isinstance(nxt, NavigableString)
                                or not soup.has_any_classes(nxt, *self.__compound_classes, r'o', r'p')
                                or not self.__ns_token_expr.fullmatch(nxt.string)
                            ):
                                break
                            current = nxt
                            tags.append(current)
                            compound_name_evaluated_tags.add(id(current))

                        full_str = ''.join([tag.get_text() for tag in tags])
                        if self.__ns_full_expr.fullmatch(full_str):
                            while tags and tags[0].string == '::':
                                del tags[0]
                            while tags and tags[-1].string == '::':
                                del tags[-1]
                            if tags:
                                compound_names.append(tags)

                    # types, namespaces, enums, free functions
                    for tags in compound_names:
                        if self.__colourize_compound_def(tags, context):
                            changed_this_block = True

                # functions:
                if 1:
                    spans = code_block(r'span', class_=(r'n', r'nc'), string=True)
                    for func in spans:
                        if not self.__func_name.fullmatch(func.string):
                            continue
                        bracket = func.next_sibling
                        if (
                            bracket is None  #
                            or isinstance(bracket, NavigableString)  #
                            or r'p' not in soup.get_classes(bracket)  #
                            or not self.__func_bracket.search(bracket.string)
                        ):
                            continue
                        soup.set_class(func, r'nf')
                        changed_this_block = True

                # keywords
                spans = code_block(r'span', class_=self.__compound_classes, string=True)
                for span in spans:
                    if span.string in self.__keywords:
                        soup.set_class(span, r'k')  # Keyword
                        changed_this_block = True

                if changed_this_block:
                    code_block.smooth()
                    changed_this_pass = True
            changed = changed or changed_this_pass

        # fix doxygen butchering code blocks as inline nonsense
        code_blocks = doc.body('code', class_=('m-code', 'm-console'))
        changed_this_pass = True
        while changed_this_pass:
            changed_this_pass = False
            for code_block in code_blocks:
                parent = code_block.parent
                if (
                    parent is None
                    or parent.name != r'p'
                    or parent.parent is None
                    or parent.parent.name not in (r'div', r'section')
                ):
                    continue
                changed_this_pass = True
                code_block.name = 'pre'
                parent.insert_before(code_block.extract())
                parent.smooth()
                if not parent.contents or (len(parent.contents) == 1 and parent.contents[0].string.strip() == ''):
                    soup.destroy_node(parent)
            changed = changed or changed_this_pass

        return changed


class AutoDocLinks(HTMLFixer):
    '''
    Adds links to additional sources where appropriate.
    '''

    __allowedNames = ('dd', 'p', 'dt', 'h3', 'td', 'div', 'figcaption')

    @classmethod
    def __substitute(cls, m, uri):
        external = uri.startswith('http')
        return rf'''<a href="{uri}" class="m-doc poxy-injected{' poxy-external' if external else ''}"{' target="_blank"' if external else ''}>{m[0]}</a>'''

    def __call__(self, context: Context, doc: soup.HTMLDocument, path: Path):
        if doc.article_content is None:
            return False

        changed = False

        # first check all existing doc links to make sure they aren't erroneously linked to the wrong thing
        if 1:

            def m_doc_anchor_tags(tag):
                return (
                    tag.name == 'a'
                    and tag.has_attr('class')
                    and ('m-doc' in tag['class'])  # or 'm-doc-self' in tag['class'])
                    and (tag.string is not None or tag.strings is not None)
                )

            existing_doc_links = doc.article_content.find_all(m_doc_anchor_tags)
            for link in existing_doc_links:
                done = False
                s = link.get_text()
                for expr, uri in context.autolinks:
                    # check that it's a match for the replacement expression
                    if not expr.fullmatch(s):
                        continue
                    # check the existing href against the target first
                    if link.has_attr('href'):
                        href = str(link['href'])
                        anchor = href.rfind('#')
                        if anchor == 0:
                            continue  # don't override internal self-links
                        if anchor != -1:
                            href = href[:anchor]
                        if href == uri or href == path.name:  # don't override internal self-links
                            continue
                    link['href'] = uri
                    soup.set_class(link, ['m-doc', 'poxy-injected'])
                    if uri.startswith('http'):
                        soup.add_class(link, 'poxy-external')
                    done = True
                    changed = True
                    break
                if done:
                    continue

        # now search the document for any other potential links
        if 1:
            tags = soup.shallow_search(
                doc.article_content,
                self.__allowedNames,
                lambda t: soup.find_parent(t, 'a', doc.article_content) is None,
            )
            strings = []
            for tag in tags:
                strings = strings + soup.string_descendants(tag, lambda t: soup.find_parent(t, 'a', tag) is None)
            strings = [s for s in strings if s.parent is not None]
            for expr, uri in context.autolinks:
                if uri == path.name:  # don't create unnecessary self-links
                    continue
                i = 0
                while i < len(strings):
                    string = strings[i]
                    parent = string.parent
                    replacer = RegexReplacer(
                        expr, lambda m, out: self.__substitute(m, uri), html.escape(str(string), quote=False)
                    )
                    if replacer:
                        repl_str = str(replacer)
                        begins_with_ws = len(repl_str) > 0 and repl_str[:1].isspace()
                        new_tags = soup.replace_tag(string, repl_str)
                        if begins_with_ws and new_tags[0].string is not None and not new_tags[0].string[:1].isspace():
                            new_tags[0].insert_before(' ')
                        changed = True
                        del strings[i]
                        for tag in new_tags:
                            strings = strings + soup.string_descendants(
                                tag, lambda t: soup.find_parent(t, 'a', parent) is None
                            )
                        continue
                    i = i + 1
        return changed


class Links(HTMLFixer):
    '''
    Fixes various minor issues with anchor tags.
    '''

    __external_href = re.compile(r'^(?:https?|s?ftp|mailto)[:].+$', re.I)
    __internal_doc_id = re.compile(r'^[a-fA-F0-9]+$')
    __local_href = re.compile(r'^([-/_a-zA-Z0-9]+\.[a-zA-Z]+)(?:#(.*))?$')
    __godbolt = re.compile(r'^\s*(?:https?[:]//)?(?:www[.])?godbolt[.]org/z/.+?$', re.I)
    __cppreference = re.compile(r'^\s*(?:https?[:]//)?(?:[a-z]+[.])?cppreference[.]com/.*$', re.I)
    __named_req = re.compile(r'^\s*(?:https?[:]//)?(?:[a-z]+[.])?cppreference[.]com/.+?/named_req/.+?$', re.I)

    def __call__(self, context: Context, doc: soup.HTMLDocument, path: Path):
        changed = False

        elems_with_ids = [e for e in doc.body(id=True) if e['id'] is not None and len(e['id'])]
        elems_with_ids = {e['id']: e for e in elems_with_ids}

        for anchor in doc.body('a', href=True):
            # make sure internal links to #ids on the same page don't get treated as external links
            # (some versions of doxygen did this with @ref)
            if anchor['href'].startswith(rf'{path.name}#'):
                anchor['href'] = anchor['href'][len(rf'{path.name}') :]
                changed = True

            # tag links to cppreference.com
            if self.__cppreference.fullmatch(anchor['href']):
                changed = soup.add_class(anchor, 'poxy-cppreference') or changed

            # tag links to cpp named requirements
            if self.__named_req.fullmatch(anchor['href']):
                changed = soup.add_class(anchor, 'poxy-named-requirement') or changed

            # make sure links to external sources are correctly marked as such
            if self.__external_href.fullmatch(anchor['href']) is not None:
                if 'target' not in anchor.attrs or anchor['target'] != '_blank':
                    anchor['target'] = '_blank'
                    changed = True
                changed = soup.add_class(anchor, 'poxy-external') or changed

                # do magic with godbolt.org links
                if self.__godbolt.fullmatch(anchor['href']):
                    changed = soup.add_class(anchor, 'poxy-godbolt') or changed
                    if (
                        anchor.parent.name == 'p'
                        and len(anchor.parent.contents) == 1
                        and anchor.parent.next_sibling is not None
                        and anchor.parent.next_sibling.name in ('pre', 'code')
                    ):
                        soup.add_class(anchor.parent, ('m-note', 'm-success', 'poxy-godbolt'))
                        code_block = anchor.parent.next_sibling
                        code_block.insert(0, anchor.parent.extract())
                        changed = True
                continue

            is_mdoc = r'class' in anchor.attrs and (r'm-doc' in anchor['class'] or r'm-doc-self' in anchor['class'])

            # make sure links to local files point to actual existing files
            match = self.__local_href.fullmatch(anchor['href'])
            if match and not coerce_path(path.parent, match[1]).exists():
                changed = True
                # fix for some doxygen versions not emitting the 'md_' prefix:
                if match[1].startswith(r'md_'):
                    repl_name = match[1][3:]
                    if repl_name and coerce_path(path.parent, repl_name).exists():
                        anchor[r'href'] = repl_name
                        continue
                # non-existent hrefs that correspond to internal documentation can sometimes by fixed by the next step
                if is_mdoc:
                    anchor['href'] = r'#'
                # otherwise this is a href to a non-existent file so we just convert it to a plain span
                else:
                    for attr in (
                        r'download',
                        r'href',
                        r'hreflang',
                        r'media',
                        r'ping',
                        r'referrerpolicy',
                        r'rel',
                        r'target',
                        r'type',
                    ):
                        if attr in anchor.attrs:
                            del anchor[attr]
                    anchor.name = r'span'
                    soup.add_class(anchor, 'poxy-dead-link')
                    continue

            # make sure internal documentation #id links actually have somewhere to go
            if (
                is_mdoc
                and anchor['href'].startswith(r'#')
                and (len(anchor['href']) == 1 or anchor['href'][1:] not in elems_with_ids)
            ):
                changed = True
                soup.remove_class(anchor, 'm-doc')
                soup.add_class(anchor, 'm-doc-self')
                anchor['href'] = '#'
                parent_with_id = anchor.find_parent(id=self.__internal_doc_id, cutoff=doc.body)
                if parent_with_id is None:
                    parent_with_id = anchor.find_parent((r'dt', r'tr'), id=False, cutoff=doc.body)
                    if parent_with_id is not None:
                        parent_with_id['id'] = sha256(parent_with_id.get_text())
                        elems_with_ids[parent_with_id['id']] = parent_with_id
                if parent_with_id is None:
                    parent_with_id = anchor.find_parent(id=True, cutoff=doc.body)
                if parent_with_id is not None:
                    anchor['href'] = '#' + parent_with_id['id']
                continue

        return changed


class EmptyTags(HTMLFixer):
    '''
    Prunes the tree of various empty tags (happens as a side-effect of some other operations).
    '''

    def __call__(self, context: Context, doc: soup.HTMLDocument, path: Path):
        changed = False
        for tag in doc.body((r'p', r'span')):
            if not tag.contents or (
                len(tag.contents) == 1 and isinstance(tag.contents[0], NavigableString) and not tag.string
            ):
                soup.destroy_node(tag)
                changed = True
        return changed


class MarkTOC(HTMLFixer):
    '''
    Marks any table-of-contents with a custom class.
    '''

    def __call__(self, context: Context, doc: soup.HTMLDocument, path: Path):
        if doc.table_of_contents is None:
            return False
        soup.add_class(doc.table_of_contents, r'poxy-toc')
        doc.table_of_contents['id'] = r'poxy-toc'
        soup.add_class(doc.body, r'poxy-has-toc')
        return True


class InjectSVGs(HTMLFixer):
    '''
    Injects the contents of SVG <img> tags directly into the document.
    '''

    def __call__(self, context: Context, doc: soup.HTMLDocument, path: Path):
        imgs = doc.body.find_all(r'img')
        if not imgs:
            return False
        imgs = [
            i
            for i in imgs
            if r'src' in i.attrs and i[r'src'] and not is_uri(i[r'src']) and i[r'src'].lower().endswith(r'.svg')
        ]
        count = 0
        for img in imgs:
            src = Path(path.parent, img[r'src'])
            if not src.exists() or not src.is_file() or src.stat().st_size > (1024 * 16):  # max 16 kb
                continue
            svg = SVG(
                src,  #
                logger=context.verbose_logger,
                root_id=img[r'id'] if r'id' in img.attrs else rf'poxy-injected-svg-{count}',
                root_classes=(*soup.get_classes(img), r'poxy-injected-svg'),
            )
            img = soup.replace_tag(img, str(svg))[0]
            count += 1
        return count > 0


class RemoveTemplateNoise(HTMLFixer):
    '''
    Removes some template noise from detail blocks.
    '''

    def __call__(self, context: Context, doc: soup.HTMLDocument, path: Path):
        tags = [tag for tag in doc.article.find_all(r'span', class_='m-doc-details-prefix') if not tag.decomposed]
        changed = False
        for tag in tags:
            m = re.fullmatch(r'([a-zA-Z_][a-zA-Z_0-9:]*)<.+?>::', tag.get_text())
            if not m:
                continue
            tag.string = rf'{m[1]}::'
            changed = True
        return changed


# =======================================================================================================================
# plain text post-processes
# =======================================================================================================================


class ImplementationDetails(PlainTextFixer):
    '''
    Replaces implementation details with appropriate shorthands.
    '''

    __shorthands = ((r'POXY_IMPLEMENTATION_DETAIL_IMPL', r'<code class="m-note m-dim poxy-impl">/* ... */</code>'),)

    def __call__(self, context: Context, text: str, path: Path) -> str:
        for shorthand, replacement in self.__shorthands:
            idx = text.find(shorthand)
            while idx >= 0:
                text = text[:idx] + replacement + text[idx + len(shorthand) :]
                idx = text.find(shorthand)
        return text


WBR = r'(?:<wbr[ \t]*/?>)?'


class MarkdownPages(PlainTextFixer):
    '''
    Cleans up some HTML snafus from markdown-based pages.
    '''

    def __call__(self, context: Context, text: str, path: Path) -> str:
        lower_name = path.name.lower()
        if (
            lower_name.startswith(r'md_')  #
            or lower_name.startswith(r'm_d__')  #
            or (context.changelog and lower_name == r'poxy_changelog.html')
        ):
            PREFIX = rf'_{WBR}_{WBR}poxy_{WBR}thiswasan_{WBR}'
            text = re.sub(rf'{PREFIX}amp', r'&amp;', text)
            text = re.sub(rf'{PREFIX}at', r'@', text)
            text = re.sub(rf'{PREFIX}fe0f', r'&#xFE0F;', text)
        return text


class ReturnTypes(PlainTextFixer):
    '''
    Fixes various issues with function return types
    '''

    __deduced_auto_return_type_brief = re.compile(
        rf'\)[ \t]*-&gt;[ \t]*_{WBR}_{WBR}poxy_{WBR}deduced_{WBR}auto_{WBR}return_{WBR}type'
    )
    __deduced_auto_return_type = re.compile(rf'_{WBR}_{WBR}poxy_{WBR}deduced_{WBR}auto_{WBR}return_{WBR}type')

    def __call__(self, context: Context, text: str, path: Path) -> str:

        # fix 'auto() -> auto'
        text = self.__deduced_auto_return_type_brief.sub(r')', text)
        text = self.__deduced_auto_return_type.sub(r'auto', text)

        # fix missing whitespace between qualifiers
        text = re.sub(
            r'(<span\s+class="m-doc-wrap-bumper"\s*>)(const|volatile|const\s+volatile|volatile\s+const)(<a\s+class="m-doc")',
            r'\1\2&nbsp;\3',
            text,
        )
        text = re.sub(
            r'\)((?:\s+(?:const|volatile|mutable|noexcept|&|&&|&amp;|&amp;&amp;))*)\s*-&gt;\s*(const|volatile|const\s+volatile|volatile\s+const)(<a\s+class="m-doc")',
            r')\1&nbsp;&rarr;&nbsp;\2&nbsp;\3',
            text,
        )

        # turn -> into &rarr;
        text = re.sub(
            r'\)((?:\s+(?:const|volatile|mutable|noexcept|&|&&|&amp;|&amp;&amp;))*)\s+-&gt;',
            r')\1&nbsp;&rarr;&nbsp;',
            text,
        )

        return text


BUILTIN_LITERALS = TrieRegEx()
for l in (r'l', r'L'):
    BUILTIN_LITERALS.add(l)  # long
    for l2 in (r'l', r'L'):
        BUILTIN_LITERALS.add(rf'{l}{l2}')  # long long
        for u in (r'u', r'U'):
            BUILTIN_LITERALS.add(rf'{u}{l}{l2}')  # unsigned long long
    for u in (r'u', r'U'):
        BUILTIN_LITERALS.add(rf'{u}{l}')  # unsigned long
for f in (r'f', r'F'):
    BUILTIN_LITERALS.add(f)  # float
for q in (r'q', r'Q'):
    BUILTIN_LITERALS.add(f)  # quad
# std::chrono
BUILTIN_LITERALS.add(r'd')
BUILTIN_LITERALS.add(r'h')
BUILTIN_LITERALS.add(r'min')
BUILTIN_LITERALS.add(r'ms')
BUILTIN_LITERALS.add(r'ns')
BUILTIN_LITERALS.add(r's')
BUILTIN_LITERALS.add(r'us')
BUILTIN_LITERALS.add(r'y')
# std::complex
BUILTIN_LITERALS.add(r'i')
BUILTIN_LITERALS.add(r'if')
BUILTIN_LITERALS.add(r'il')
# std::string(view)
BUILTIN_LITERALS.add(r's')
BUILTIN_LITERALS.add(r'sv')
BUILTIN_LITERALS = BUILTIN_LITERALS.regex()


class Pygments(PlainTextFixer):
    '''
    Fixes minor issues with pygments-generated markup.
    '''

    def __call__(self, context: Context, text: str, path: Path) -> str:
        if not re.search(r'class="[^"]*?m-code[^"]*?"', text):
            return None

        # at some point pygments started adding markup to whitespace,
        # causing an awful lot of markup bloat. m.css does not style this markup
        # so we can safely strip it away.
        text = re.sub(r'<span class="w">(\s+)</span>', r'\1', text)

        # fix numeric UDLs being treated as a separate token
        text = re.sub(
            rf'<span\s+class="(m[bfhio])"\s*>(.*?)</span><span class="n">((?:_[a-zA-Z0-9_]*)|{BUILTIN_LITERALS})</span>',  #
            r'<span class="\1">\2\3</span>',
            text,
        )

        # fix string UDLs being treated as a separate token
        text = re.sub(
            rf'<span\s+class="s"\s*>(.*?)</span><span class="n">((?:_[a-zA-Z0-9_]*)|{BUILTIN_LITERALS})</span>',  #
            r'<span class="s">\1\2</span>',
            text,
        )

        # hack to make some basic #ifs, #defines etc. look nice
        text = re.sub(
            r'<span\s+class="cp"\s*>(\s*#\s*(?:(?:el)?if(?:n?def)?|define|undef)\s+)([a-zA-Z_][a-zA-Z_0-9]*?)([^a-zA-Z_0-9])',  #
            r'<span class="cp">\1</span><span class="fm">\2</span><span class="cp">\3',
            text,
        )

        # hack to make basic "using XXXX = ..." directives look nice
        text = re.sub(
            r'<span\s+class="k"\s*>(\s*using\s*)</span>(\s+)<span\s+class="n"\s*>([a-zA-Z_][a-zA-Z0-9_]*?)</span>(\s+)<span\s+class="o"\s*>(\s*=\s*)</span>',
            r'<span class="k">\1</span>\2<span class="nc">\3</span>\4<span class="o">\5</span>',
            text,
        )

        return text


class InstallSearchShim(PlainTextFixer):
    '''
    Installs our shim around m.css' showSearch().
    '''

    def __call__(self, context: Context, text: str, path: Path) -> str:
        return re.sub(
            r'<\s*script\s+src="search-v2[.]js"\s*>\s*</script>',
            r'<script src="search-v2.js"></script><script>install_mcss_search_shim();</script>',
            text,
            flags=re.DOTALL,
        )


def create_all() -> Tuple[Union[HTMLFixer, PlainTextFixer]]:

    # order matters here!
    return (
        MarkTOC(),  # html
        Pygments(),
        CodeBlocks(),  # html
        Banner(),  # html
        CPPModifiers1(),  # html
        CPPModifiers2(),  # html
        StripIncludes(),  # html
        AutoDocLinks(),  # html
        Links(),  # html
        CustomTags(),  # html
        RemoveTemplateNoise(),  # html
        EmptyTags(),  # html
        ImplementationDetails(),
        MarkdownPages(),
        InstallSearchShim(),
        ReturnTypes(),
        InjectSVGs(),  # html
    )


__all__ = ['HTMLFixer', 'PlainTextFixer', 'create_all']



#!/usr/bin/env python3
"""latex2svg

Read LaTeX code from stdin and render a SVG using LaTeX + dvisvgm.
"""
__version__ = '0.1.0'
__author__ = 'Tino Wagner'
__email__ = 'ich@tinowagner.com'
__license__ = 'MIT'
__copyright__ = '(c) 2017, Tino Wagner'

import os
import sys
import subprocess
import shlex
import re
from tempfile import TemporaryDirectory
from ctypes.util import find_library

default_template = r"""
\documentclass[{{ fontsize }}pt,preview]{standalone}
{{ preamble }}
\begin{document}
\begin{preview}
{{ code }}
\end{preview}
\end{document}
"""

default_preamble = r"""
\usepackage[utf8x]{inputenc}
\usepackage{amsmath}
\usepackage{amsfonts}
\usepackage{amssymb}
\usepackage{newtxtext}
\usepackage[libertine]{newtxmath}
"""

latex_cmd = 'latex -interaction nonstopmode -halt-on-error'
dvisvgm_cmd = 'dvisvgm --no-fonts'

default_params = {
    'fontsize': 12,  # pt
    'template': default_template,
    'preamble': default_preamble,
    'latex_cmd': latex_cmd,
    'dvisvgm_cmd': dvisvgm_cmd,
    'libgs': None,
}

libgs = find_library('gs')
if not hasattr(os.environ, 'LIBGS') and not libgs:
    if sys.platform == 'darwin':
        # Fallback to homebrew Ghostscript on macOS
        homebrew_libgs = '/usr/local/opt/ghostscript/lib/libgs.dylib'
        if os.path.exists(homebrew_libgs):
            default_params['libgs'] = homebrew_libgs
    if not default_params['libgs']:
        print('Warning: libgs not found')
# dvisvgm < 3.0 only looks for ghostscript < 10 on its own, attempt to supply
# it directly if version 10 is found. Fixed in
# https://github.com/mgieseki/dvisvgm/commit/46b11c02a46883309a824e3fc798f8093041daec
# TODO related https://bugs.archlinux.org/task/76083 was resolved on May 29th
# 2023, remove after enough time passes if no similar case happens in other
# distros
elif '.so.10' in str(libgs):
    # If dvisvgm seems to be at least 3.0, don't even attempt to find libgs and
    # just assume it works
    dvisvgm_is_new_enough = False
    try:
        ret = subprocess.run(['dvisvgm', '--version'], stdout=subprocess.PIPE)
        if b'dvisvgm 3' in ret.stdout:
            dvisvgm_is_new_enough = True
    except:
        pass

    # Just winging it here, there doesn't seem to be an easy way to get the
    # actual full path the library was found in -- ctypes/util.py does crazy
    # stuff like invoking gcc (!!) to get the library filename.
    #
    # On Arch /usr/lib64 is a symlink to /usr/lib, so only the former is
    # needed; on Fedora it's two distinct directories and libgs is in the
    # latter (ugh). Elsewhere it might be /usr/lib/x86_64-linux-gnu or just
    # anything else, so let's just bet that dvisvgm is 3.0+ in most cases and
    # this gets hit only very rarely.
    if not dvisvgm_is_new_enough:
        prefixes = ['/usr/lib', '/usr/lib64']
        for prefix in prefixes:
            libgs_absolute = os.path.join(prefix, libgs)
            if os.path.exists(libgs_absolute):
                default_params['libgs'] = libgs_absolute
                break
        else:
            raise RuntimeError('libgs found by linker magic, but is not in {}'.format(' or '.join(prefixes)))

def latex2svg(code, params=default_params, working_directory=None):
    """Convert LaTeX to SVG using dvisvgm.

    Parameters
    ----------
    code : str
        LaTeX code to render.
    params : dict
        Conversion parameters.
    working_directory : str or None
        Working directory for external commands and place for temporary files.

    Returns
    -------
    dict
        Dictionary of SVG output and output information:

        * `svg`: SVG data
        * `width`: image width in *em*
        * `height`: image height in *em*
        * `depth`: baseline position in *em*
    """
    if working_directory is None:
        with TemporaryDirectory() as tmpdir:
            return latex2svg(code, params, working_directory=tmpdir)

    fontsize = params['fontsize']
    document = (params['template']
                .replace('{{ preamble }}', params['preamble'])
                .replace('{{ fontsize }}', str(fontsize))
                .replace('{{ code }}', code))

    with open(os.path.join(working_directory, 'code.tex'), 'w') as f:
        f.write(document)

    # Run LaTeX and create DVI file
    try:
        ret = subprocess.run(shlex.split(params['latex_cmd']+' code.tex'),
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                             cwd=working_directory)
        # LaTeX prints errors on stdout instead of stderr (stderr is empty),
        # so print stdout instead
        if ret.returncode: print(ret.stdout.decode('utf-8'))
        ret.check_returncode()
    except FileNotFoundError:
        raise RuntimeError('latex not found')

    # Add LIBGS to environment if supplied
    env = os.environ.copy()
    if params['libgs']:
        env['LIBGS'] = params['libgs']

    # Convert DVI to SVG
    try:
        ret = subprocess.run(shlex.split(params['dvisvgm_cmd']+' code.dvi'),
                             stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                             cwd=working_directory, env=env)
        if ret.returncode: print(ret.stderr.decode('utf-8'))
        ret.check_returncode()
    except FileNotFoundError:
        raise RuntimeError('dvisvgm not found')

    with open(os.path.join(working_directory, 'code.svg'), 'r') as f:
        svg = f.read()

    # Parse dvisvgm output for size and alignment
    def get_size(output):
        regex = r'\b([0-9.]+)pt x ([0-9.]+)pt'
        match = re.search(regex, output)
        if match:
            return (float(match.group(1)) / fontsize,
                    float(match.group(2)) / fontsize)
        else:
            return None, None

    def get_measure(output, name):
        regex = r'\b%s=([0-9.e-]+)pt' % name
        match = re.search(regex, output)
        if match:
            return float(match.group(1)) / fontsize
        else:
            return None

    output = ret.stderr.decode('utf-8')

    # Since we rely on Ghostscript to give us a "depth" (= vertical alignment)
    # value, dvisvgm not using it means the output would be broken and it makes
    # no sense to continue.
    if 'Ghostscript not found' in output:
        raise RuntimeError('libgs not detected by dvisvgm, point the LIBGS environment variable to its location')

    width, height = get_size(output)
    depth = get_measure(output, 'depth')
    return {'svg': svg, 'depth': depth, 'width': width, 'height': height}


def main():
    """Simple command line interface to latex2svg.

    - Read from `stdin`.
    - Write SVG to `stdout`.
    - Write metadata as JSON to `stderr`.
    - On error: write error messages to `stdout` and return with error code.
    """
    import json
    import argparse
    parser = argparse.ArgumentParser(description="""
    Render LaTeX code from stdin as SVG to stdout. Writes metadata (baseline
    position, width, height in em units) as JSON to stderr.
    """)
    parser.add_argument('--preamble',
                        help="LaTeX preamble code to read from file")
    args = parser.parse_args()
    preamble = default_preamble
    if args.preamble is not None:
        with open(args.preamble) as f:
            preamble = f.read()
    latex = sys.stdin.read()
    try:
        params = default_params.copy()
        params['preamble'] = preamble
        out = latex2svg(latex, params)
        sys.stdout.write(out['svg'])
        meta = {key: out[key] for key in out if key != 'svg'}
        sys.stderr.write(json.dumps(meta))
    except subprocess.CalledProcessError as exc:
        print(exc.output.decode('utf-8'))
        sys.exit(exc.returncode)


if __name__ == '__main__':
    main()
